import asyncio
from json import JSONDecodeError
from typing import List, Optional
from unittest.mock import patch

import marshmallow
import pytest

from starknet_py.devnet_utils.devnet_client import DevnetClient
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client_models import (
    BlockHeader,
    Call,
    EmittedEventWithFinalityStatus,
    StarknetBlock,
    TransactionFinalityStatus,
    TransactionFinalityStatusWithoutL1,
    TransactionReceipt,
    TransactionStatusWithoutL1,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.websockets.errors import WebsocketClientError
from starknet_py.net.websockets.models import (
    NewEventsNotification,
    NewHeadsNotification,
    NewTransactionNotification,
    NewTransactionNotificationResult,
    NewTransactionReceiptsNotification,
)
from starknet_py.net.websockets.websocket_client import WebsocketClient
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_subscribe_new_heads_with_block_hash(
    client: FullNodeClient,
    websocket_client: WebsocketClient,
    devnet_client: DevnetClient,
):
    received_block: Optional[BlockHeader] = None

    def handler(new_heads_notification: NewHeadsNotification):
        nonlocal received_block
        received_block = new_heads_notification.result

    latest_block = await client.get_block(block_hash="latest")
    assert isinstance(latest_block, StarknetBlock)

    subscription_id = await websocket_client.subscribe_new_heads(
        handler=handler, block_hash=latest_block.block_hash
    )

    new_block_hash = await devnet_client.create_block()

    assert received_block is not None
    assert int(new_block_hash, 16) == received_block.block_hash

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.asyncio
async def test_subscribe_new_heads_with_block_number(
    client: FullNodeClient,
    websocket_client: WebsocketClient,
    devnet_client: DevnetClient,
):
    received_block: Optional[BlockHeader] = None

    def handler(new_heads_notification: NewHeadsNotification):
        nonlocal received_block
        received_block = new_heads_notification.result

    latest_block = await client.get_block(block_hash="latest")
    assert isinstance(latest_block, StarknetBlock)

    subscription_id = await websocket_client.subscribe_new_heads(
        handler=handler, block_number=latest_block.block_number
    )

    new_block_hash = await devnet_client.create_block()

    assert received_block is not None
    assert int(new_block_hash, 16) == received_block.block_hash

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.asyncio
async def test_subscribe_new_heads_with_both_block_params_passed(
    websocket_client: WebsocketClient,
):
    with pytest.raises(
        ValueError,
        match="Arguments block_hash and block_number are mutually exclusive.",
    ):
        await websocket_client.subscribe_new_heads(
            handler=lambda _: _, block_hash=1, block_number=1
        )


@pytest.mark.asyncio
async def test_subscribe_new_heads_too_many_blocks_back(
    websocket_client: WebsocketClient,
    devnet_client_fork_mode: DevnetClient,
):
    client = FullNodeClient(devnet_client_fork_mode.url)
    latest_block = await client.get_block(block_hash="latest")

    assert isinstance(latest_block, StarknetBlock)
    assert latest_block.block_number >= 1025

    # TODO(#1574): Change error to `TOO_MANY_BLOCKS_BACK` once devnet issue is resolved.
    query_block_number = latest_block.block_number - 1025
    with pytest.raises(
        WebsocketClientError,
        match="WebsocketClient failed with code 24. Message: Block not found.",
    ):
        await websocket_client.subscribe_new_heads(
            handler=lambda _: _, block_number=query_block_number
        )


@pytest.mark.asyncio
async def test_unsubscribe_with_non_existing_id(
    websocket_client: WebsocketClient,
):
    unsubscribe_result = await websocket_client.unsubscribe("123")
    assert unsubscribe_result is False


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_events_with_finality_status(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    received_events: List = []

    def handler(new_events_notification: NewEventsNotification):
        nonlocal received_events
        received_events.append(new_events_notification.result)

    subscription_id = await websocket_client.subscribe_events(
        handler=handler,
        # We subscribe to the events from the account because it emits them for each executed transaction.
        # Balance contract doesn't emit anything.
        from_address=argent_account_v040.address,
        finality_status=TransactionFinalityStatusWithoutL1.ACCEPTED_ON_L2,
    )

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await asyncio.sleep(3)

    assert len(received_events) >= 1
    assert all(
        ev.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2
        for ev in received_events
    )

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_new_transactions_with_finality_status(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    received: List[NewTransactionNotificationResult] = []

    def handler(new_tx_notification: NewTransactionNotification):
        nonlocal received
        received.append(new_tx_notification.result)

    subscription_id = await websocket_client.subscribe_new_transactions(
        handler=handler,
        sender_address=[argent_account_v040.address],
        finality_status=[TransactionStatusWithoutL1.ACCEPTED_ON_L2],
    )

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await asyncio.sleep(3)

    assert len(received) >= 1
    assert all(
        r.finality_status == TransactionStatusWithoutL1.ACCEPTED_ON_L2 for r in received
    )

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_new_transaction_receipts_with_finality_status(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    receipts: List[TransactionReceipt] = []

    def handler(new_tx_receipt: NewTransactionReceiptsNotification):
        nonlocal receipts
        receipts.append(new_tx_receipt.result)

    subscription_id = await websocket_client.subscribe_new_transaction_receipts(
        handler=handler,
        sender_address=[argent_account_v040.address],
        finality_status=[TransactionFinalityStatusWithoutL1.ACCEPTED_ON_L2],
    )

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await asyncio.sleep(10)

    assert len(receipts) >= 1
    assert all(
        r.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2 for r in receipts
    )

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_events_with_all_filters(
    client: FullNodeClient,
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    received_events: List[EmittedEventWithFinalityStatus] = []

    def handler(new_events_notification: NewEventsNotification):
        nonlocal received_events
        received_events.append(new_events_notification.result)

    latest_block = await client.get_block(block_hash="latest")
    assert isinstance(latest_block, StarknetBlock)

    subscription_id = await websocket_client.subscribe_events(
        handler=handler,
        from_address=argent_account_v040.address,
        keys=[[]],
        block_number=latest_block.block_number,
        finality_status=TransactionFinalityStatusWithoutL1.ACCEPTED_ON_L2,
    )

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await asyncio.sleep(3)

    assert len(received_events) >= 1
    assert all(
        ev.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2
        for ev in received_events
    )
    assert received_events[0].from_address == argent_account_v040.address

    execute_receipt = await client.get_transaction_receipt(execute.transaction_hash)
    assert received_events[0].block_number is not None
    assert received_events[0].block_number + 1 == execute_receipt.block_number

    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    assert unsubscribe_result is True


@pytest.mark.asyncio
async def test_subscribe_failure():
    # pylint: disable=no-self-use
    class FakeConnection:
        async def recv(self):
            return "xyz"

        # pylint: disable=unused-argument
        async def send(self, *args, **kwargs):
            return

        async def close(self):
            return

    async def fake_connect(*_args, **_kwargs):
        return FakeConnection()

    with patch(
        "starknet_py.net.websockets.websocket_client.connect", side_effect=fake_connect
    ):
        ws = WebsocketClient("wss://example.invalid")
        await ws.connect()

        with pytest.raises(JSONDecodeError):
            await ws.subscribe_events(lambda _: None)

        await ws.disconnect()


@pytest.mark.asyncio
async def test_listener_failure():
    # pylint: disable=no-self-use
    class FakeConnection:
        async def recv(self):
            # pylint: disable=line-too-long
            return '{"method": "starknet_subscriptionNewTransactionReceipts","params": {"subscription_id": "1234", "result": {"unknown_key": 12345}}}'

        async def send(self):
            return

        async def close(self):
            return

    async def fake_connect(*_args, **_kwargs):
        return FakeConnection()

    with patch(
        "starknet_py.net.websockets.websocket_client.connect", side_effect=fake_connect
    ):
        ws = WebsocketClient("wss://example.invalid")
        await ws.connect()

        # pylint: disable=protected-access
        ws._subscriptions["1234"] = lambda _: None

        with pytest.raises(marshmallow.ValidationError):
            await ws.wait_closed_or_failed()

        await ws.disconnect()



# pylint: disable=import-outside-toplevel, duplicate-code
import os

import pytest

directory = os.path.dirname(__file__)


@pytest.mark.asyncio
async def test_using_contract(account, map_contract):
    # pylint: disable=unused-variable,too-many-locals
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

    contract_address = (
        "0x01336fa7c870a7403aced14dda865b75f29113230ed84e3a661f7af70fe83e7b"
    )
    key = 1234
    # docs: end

    contract_address = map_contract.address
    # docs: start

    # Create contract from contract's address - Contract will download contract's ABI to know its interface.
    contract = await Contract.from_address(address=contract_address, provider=account)
    # docs: end

    abi = contract.data.abi

    # docs: start

    # If the ABI is known, create the contract directly (this is the preferred way).
    contract = Contract(
        address=contract_address,
        abi=abi,
        provider=account,
    )

    # All exposed functions are available at contract.functions.
    # Here we invoke a function, creating a new transaction.
    invocation = await contract.functions["put"].invoke_v3(
        key,
        7,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )

    # Invocation returns InvokeResult object. It exposes a helper for waiting until transaction is accepted.
    await invocation.wait_for_acceptance()

    # Calling contract's function doesn't create a new transaction, you get the function's result.
    (saved,) = await contract.functions["get"].call(key)
    # saved = 7 now
    # docs: end

    assert saved == 7



# pylint: disable=import-outside-toplevel
import asyncio
from typing import List, Optional

import pytest

from starknet_py.devnet_utils.devnet_client import DevnetClient
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client_models import (
    BlockHeader,
    Call,
    EmittedEventWithFinalityStatus,
    TransactionExecutionStatus,
    TransactionReceipt,
    TransactionStatus,
)
from starknet_py.net.websockets.models import NewTransactionStatus, ReorgData
from starknet_py.net.websockets.websocket_client import WebsocketClient
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_connect_and_disconnect(devnet_ws: str):
    # pylint: disable=unused-variable
    # pylint: disable=redefined-outer-name
    # pylint: disable=reimported
    # docs-start: connect
    from starknet_py.net.websockets.websocket_client import WebsocketClient

    # Create `WebsocketClient` instance
    websocket_client = WebsocketClient("wss://your.node.url")
    # docs-end: connect
    websocket_client = WebsocketClient(devnet_ws)
    assert not await websocket_client.is_connected

    # docs-start: connect

    # Connect to the websocket server
    await websocket_client.connect()
    # docs-end: connect
    assert await websocket_client.is_connected

    # docs-start: disconnect
    # Disconnect from the websocket server
    await websocket_client.disconnect()
    # docs-end: disconnect
    assert not await websocket_client.is_connected


@pytest.mark.asyncio
async def test_subscribe_new_heads(
    websocket_client: WebsocketClient,
    devnet_client: DevnetClient,
):
    received_block: Optional[BlockHeader] = None

    # docs-start: subscribe_new_heads
    from starknet_py.net.websockets.models import NewHeadsNotification

    # Create a handler function that will be called when a new block is created
    def handler(new_heads_notification: NewHeadsNotification):
        # Perform the necessary actions with the new block...

        # docs-end: subscribe_new_heads
        nonlocal received_block
        received_block = new_heads_notification.result

    # docs-start: subscribe_new_heads
    # Subscribe to new heads notifications
    subscription_id = await websocket_client.subscribe_new_heads(handler=handler)

    # Here you can put code which will keep the application running (e.g. using loop and `asyncio.sleep`)
    # ...
    # docs-end: subscribe_new_heads
    new_block_hash = await devnet_client.create_block()

    assert received_block is not None
    assert int(new_block_hash, 16) == received_block.block_hash
    # docs-start: subscribe_new_heads

    # Unsubscribe from the notifications
    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    # docs-end: subscribe_new_heads
    assert unsubscribe_result is True


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_events(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    account = argent_account_v040
    emitted_events: List[EmittedEventWithFinalityStatus] = []

    # docs-start: subscribe_events
    from starknet_py.net.websockets.models import NewEventsNotification

    # Create a handler function that will be called when a new event is emitted
    def handler(new_events_notification: NewEventsNotification):
        # Perform the necessary actions with the new event...

        # docs-end: subscribe_events
        nonlocal emitted_events
        emitted_events.append(new_events_notification.result)

    # docs-start: subscribe_events
    # Subscribe to new events notifications
    subscription_id = await websocket_client.subscribe_events(
        handler=handler, from_address=account.address
    )

    # Here you can put code which will keep the application running (e.g. using loop and `asyncio.sleep`)
    # ...
    # docs-end: subscribe_events
    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await argent_account_v040.client.get_transaction_receipt(
        tx_hash=execute.transaction_hash
    )

    assert len(emitted_events) > 0
    for emitted_event in emitted_events:
        assert emitted_event.from_address == account.address
    # docs-start: subscribe_events

    # Unsubscribe from the notifications
    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    # docs-end: subscribe_events
    assert unsubscribe_result is True


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_transaction_status(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    new_transaction_status: Optional[NewTransactionStatus] = None

    # docs-start: subscribe_transaction_status
    from starknet_py.net.websockets.models import TransactionStatusNotification

    # Create a handler function that will be called when a new transaction status is emitted
    def handler(transaction_status_notification: TransactionStatusNotification):
        # Perform the necessary actions with the new transaction status...

        # docs-end: subscribe_transaction_status
        nonlocal new_transaction_status
        new_transaction_status = transaction_status_notification.result

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    # docs-start: subscribe_transaction_status
    # Subscribe to transaction status notifications
    subscription_id = await websocket_client.subscribe_transaction_status(
        handler=handler, transaction_hash=execute.transaction_hash
    )

    # Here you can put code which will keep the application running (e.g. using loop and `asyncio.sleep`)
    # ...
    # docs-end: subscribe_transaction_status
    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await argent_account_v040.client.get_transaction_receipt(
        tx_hash=execute.transaction_hash
    )

    await asyncio.sleep(5)

    assert new_transaction_status is not None
    assert new_transaction_status.transaction_hash == execute.transaction_hash
    assert (
        new_transaction_status.status.finality_status
        == TransactionStatus.ACCEPTED_ON_L2
    )
    assert (
        new_transaction_status.status.execution_status
        == TransactionExecutionStatus.SUCCEEDED
    )
    assert new_transaction_status.status.failure_reason is None

    # docs-start: subscribe_transaction_status

    # Unsubscribe from the notifications
    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    # docs-end: subscribe_transaction_status
    assert unsubscribe_result is True


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_subscribe_new_transaction_receipts(
    websocket_client: WebsocketClient,
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    account = argent_account_v040
    transaction_receipts: List[TransactionReceipt] = []

    # docs-start: subscribe_new_transaction_receipts
    from starknet_py.net.websockets.models import NewTransactionReceiptsNotification

    # Create a handler function that will be called when a new transaction is emitted
    def handler(
        new_transaction_receipts_notification: NewTransactionReceiptsNotification,
    ):
        # Perform the necessary actions with the new transaction receipts...

        # docs-end: subscribe_new_transaction_receipts
        nonlocal transaction_receipts
        transaction_receipts.append(new_transaction_receipts_notification.result)

    # docs-start: subscribe_new_transaction_receipts
    # Subscribe to new transaction receipts notifications
    subscription_id = await websocket_client.subscribe_new_transaction_receipts(
        handler=handler,
        sender_address=[account.address],
    )

    # Here you can put code which will keep the application running (e.g. using loop and `asyncio.sleep`)
    # ...
    # docs-end: subscribe_new_transaction_receipts
    # assert len(transaction_receipts) == 0

    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    await argent_account_v040.client.get_transaction_receipt(
        tx_hash=execute.transaction_hash
    )

    await asyncio.sleep(5)

    assert len(transaction_receipts) == 1
    transaction_receipt = transaction_receipts[0]

    assert execute.transaction_hash == transaction_receipt.transaction_hash

    # docs-start: subscribe_new_transaction_receipts

    # Unsubscribe from the notifications
    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    # docs-end: subscribe_new_transaction_receipts
    assert unsubscribe_result is True


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_on_chain_reorg(
    websocket_client: WebsocketClient,
    devnet_client: DevnetClient,
):
    reorg_data: Optional[ReorgData] = None

    # docs-start: on_chain_reorg
    from starknet_py.net.websockets.models import ReorgNotification

    # Create a handler function that will be called when a reorg notification is emitted
    def handler_reorg(reorg_notification: ReorgNotification):
        # Perform the necessary actions with the reorg data...

        # docs-end: on_chain_reorg
        nonlocal reorg_data
        reorg_data = reorg_notification.result

    # docs-start: on_chain_reorg
    # Set the handler function for reorg notifications
    websocket_client.on_chain_reorg = handler_reorg

    # Subscribe to new heads, events, or transaction status notifications.
    # In this example we will subscribe to new heads notifications.
    # Note: we're passing empty handler function here, as we're only interested in reorg notifications.
    subscription_id = await websocket_client.subscribe_new_heads(handler=lambda _: _)

    # Here you can put code which will keep the application running (e.g. using loop and `asyncio.sleep`)
    # ...
    # docs-end: on_chain_reorg
    new_block_hash = await devnet_client.create_block()

    await devnet_client.abort_block(block_hash=new_block_hash)

    await asyncio.sleep(5)

    assert reorg_data is not None
    assert reorg_data.starting_block_hash == int(new_block_hash, 16)
    assert reorg_data.ending_block_hash == int(new_block_hash, 16)

    # docs-start: on_chain_reorg

    # Unsubscribe from the notifications
    unsubscribe_result = await websocket_client.unsubscribe(subscription_id)
    # docs-end: on_chain_reorg
    assert unsubscribe_result is True



from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.constants import STRK_FEE_CONTRACT_ADDRESS
from starknet_py.net.account.account import Account
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import StarknetChainId, parse_address
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_get_balance_default_token_address():
    client = FullNodeClient(node_url="/rpc")
    acc_client = Account(
        client=client,
        address="0x123",
        key_pair=KeyPair(123, 456),
        chain=StarknetChainId.SEPOLIA,
    )

    with patch(
        f"{FullNodeClient.__module__}.FullNodeClient.call_contract",
        AsyncMock(),
    ) as mocked_call_contract:
        mocked_call_contract.return_value = [0, 0]

        await acc_client.get_balance()

        call = mocked_call_contract.call_args

    (call,) = call[0]

    assert call.to_addr == parse_address(STRK_FEE_CONTRACT_ADDRESS)


@pytest.mark.asyncio
async def test_account_get_balance_strk(account, hello_starknet_contract):
    balance = await account.get_balance(token_address=STRK_FEE_CONTRACT_ADDRESS)
    block = await account.client.get_block(block_number="latest")

    await hello_starknet_contract.functions["increase_balance"].invoke_v3(
        amount=10, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    new_balance = await account.get_balance(token_address=STRK_FEE_CONTRACT_ADDRESS)
    old_balance = await account.get_balance(
        token_address=STRK_FEE_CONTRACT_ADDRESS, block_number=block.block_number
    )

    assert balance > 0
    assert new_balance < balance
    assert old_balance == balance


def test_create_account():
    key_pair = KeyPair.from_private_key(0x111)
    account = Account(
        address=0x1,
        client=FullNodeClient(node_url=""),
        key_pair=key_pair,
        chain=StarknetChainId.SEPOLIA,
    )

    assert account.address == 0x1
    assert account.signer.public_key == key_pair.public_key


@pytest.mark.parametrize(
    "chain",
    [
        StarknetChainId.SEPOLIA,
        "SN_SEPOLIA",
        "0x534e5f5345504f4c4941",
        393402133025997798000961,
    ],
)
def test_create_account_parses_chain(chain):
    key_pair = KeyPair.from_private_key(0x111)
    account = Account(
        address=0x1,
        client=FullNodeClient(node_url=""),
        key_pair=key_pair,
        chain=chain,
    )

    assert account.address == 0x1
    assert account.signer.public_key == key_pair.public_key
    assert isinstance(account.signer, StarkCurveSigner)
    assert account.signer.chain_id == 0x534E5F5345504F4C4941


def test_create_account_from_signer(client):
    signer = StarkCurveSigner(
        account_address=0x1,
        key_pair=KeyPair.from_private_key(0x111),
        chain_id=StarknetChainId.SEPOLIA,
    )
    account = Account(address=0x1, client=client, signer=signer)

    assert account.address == 0x1
    assert account.signer == signer


def test_create_account_raises_on_no_keypair_and_signer():
    with pytest.raises(
        ValueError,
        match="Either a signer or a key_pair must be provided in Account constructor",
    ):
        Account(
            address=0x1,
            client=FullNodeClient(node_url=""),
            chain=StarknetChainId.SEPOLIA,
        )


def test_create_account_raises_on_both_keypair_and_signer():
    with pytest.raises(
        ValueError, match="Arguments signer and key_pair are mutually exclusive"
    ):
        Account(
            address=0x1,
            client=FullNodeClient(node_url=""),
            chain=StarknetChainId.SEPOLIA,
            key_pair=KeyPair.from_private_key(0x111),
            signer=StarkCurveSigner(
                account_address=0x1,
                key_pair=KeyPair.from_private_key(0x11),
                chain_id=StarknetChainId.SEPOLIA,
            ),
        )



from enum import IntEnum
from pathlib import Path

# Addresses came from starkware-libs/starknet-addresses repository: https://github.com/starkware-libs/starknet-addresses
ETH_FEE_CONTRACT_ADDRESS = (
    "0x049d36570d4e46f48e99674bd3fcc84644ddd6b96f7c741b1562b82f9e004dc7"
)

STRK_FEE_CONTRACT_ADDRESS = (
    "0x04718f5a0fc34cc1af16a1cdee98ffb20c31f5cd61d6ab07201858f4287c938d"
)

DEFAULT_DEPLOYER_ADDRESS = (
    "0x02ceed65a4bd731034c01113685c831b01c15d7d432f71afb1cf1634b53a2125"
)

API_VERSION = 0

RPC_CONTRACT_NOT_FOUND_ERROR = 20
RPC_INVALID_MESSAGE_SELECTOR_ERROR = 21
RPC_CLASS_HASH_NOT_FOUND_ERROR = 28
RPC_CONTRACT_ERROR = 40

DEFAULT_ENTRY_POINT_NAME = "__default__"
DEFAULT_L1_ENTRY_POINT_NAME = "__l1_default__"
DEFAULT_ENTRY_POINT_SELECTOR = 0
DEFAULT_DECLARE_SENDER_ADDRESS = 1

# MAX_STORAGE_ITEM_SIZE and ADDR_BOUND must be consistent with the corresponding constant in
# starkware/starknet/common/storage.cairo.
MAX_STORAGE_ITEM_SIZE = 256
ADDR_BOUND = 2**251 - MAX_STORAGE_ITEM_SIZE

FIELD_PRIME = 0x800000000000011000000000000000000000000000000000000000000000001
EC_ORDER = 0x800000000000010FFFFFFFFFFFFFFFFB781126DCAE7B2321E66A241ADC64D2F

# From cairo-lang
# int_from_bytes(b"STARKNET_CONTRACT_ADDRESS")
CONTRACT_ADDRESS_PREFIX = 523065374597054866729014270389667305596563390979550329787219
L2_ADDRESS_UPPER_BOUND = 2**251 - 256

QUERY_VERSION_BASE = 2**128

ROOT_PATH = Path(__file__).parent

# Result of `encode_shortstring("ANY_CALLER")`
ANY_CALLER = 0x414E595F43414C4C4552


# OUTSIDE EXECUTION INTERFACE_VERSION with ID
class OutsideExecutionInterfaceID(IntEnum):
    V1 = 0x68CFD18B92D1907B8BA3CC324900277F5A3622099431EA85DD8089255E4181
    V2 = 0x1D1144BB2138366FF28D8E9AB57456B1D332AC42196230C3A602003C89872


EXPECTED_RPC_VERSION = "0.10.0-rc.1"

ARGENT_V040_CLASS_HASH = (
    0x036078334509B514626504EDC9FB252328D1A240E4E948BEF8D0C08DFF45927F
)



from starknet_py.constants import OutsideExecutionInterfaceID
from starknet_py.net.client_models import OutsideExecution
from starknet_py.net.schemas.common import Revision
from starknet_py.utils.typed_data import TypedData

OUTSIDE_EXECUTION_INTERFACE_ID_TO_TYPED_DATA_REVISION = {
    OutsideExecutionInterfaceID.V1: Revision.V0,
    OutsideExecutionInterfaceID.V2: Revision.V1,
}


# TODO(#1537): Implement as method of OutsideExecution
def outside_execution_to_typed_data(
    outside_execution: OutsideExecution,
    outside_execution_version: OutsideExecutionInterfaceID,
    chain_id: int,
) -> TypedData:
    """
    SNIP-12 Typed Data for OutsideExecution implementation. For revision V0 and V1.
    """

    revision = OUTSIDE_EXECUTION_INTERFACE_ID_TO_TYPED_DATA_REVISION[
        outside_execution_version
    ]

    if revision == Revision.V0:
        return TypedData.from_dict(
            {
                "types": {
                    "StarkNetDomain": [
                        {"name": "name", "type": "felt"},
                        {"name": "version", "type": "felt"},
                        {"name": "chainId", "type": "felt"},
                    ],
                    "OutsideExecution": [
                        {"name": "caller", "type": "felt"},
                        {"name": "nonce", "type": "felt"},
                        {"name": "execute_after", "type": "felt"},
                        {"name": "execute_before", "type": "felt"},
                        {"name": "calls_len", "type": "felt"},
                        {"name": "calls", "type": "OutsideCall*"},
                    ],
                    "OutsideCall": [
                        {"name": "to", "type": "felt"},
                        {"name": "selector", "type": "felt"},
                        {"name": "calldata_len", "type": "felt"},
                        {"name": "calldata", "type": "felt*"},
                    ],
                },
                "primaryType": "OutsideExecution",
                "domain": {
                    "name": "Account.execute_from_outside",
                    "version": "1",
                    "chainId": str(chain_id),
                    "revision": Revision.V0,
                },
                "message": {
                    "caller": outside_execution.caller,
                    "nonce": outside_execution.nonce,
                    "execute_after": outside_execution.execute_after,
                    "execute_before": outside_execution.execute_before,
                    "calls_len": len(outside_execution.calls),
                    "calls": [
                        {
                            "to": call.to_addr,
                            "selector": call.selector,
                            "calldata_len": len(call.calldata),
                            "calldata": call.calldata,
                        }
                        for call in outside_execution.calls
                    ],
                },
            }
        )

    # revision == Revision.V1
    return TypedData.from_dict(
        {
            "types": {
                "StarknetDomain": [
                    {"name": "name", "type": "shortstring"},
                    {"name": "version", "type": "shortstring"},
                    {"name": "chainId", "type": "shortstring"},
                    {"name": "revision", "type": "shortstring"},
                ],
                "OutsideExecution": [
                    {"name": "Caller", "type": "ContractAddress"},
                    {"name": "Nonce", "type": "felt"},
                    {"name": "Execute After", "type": "u128"},
                    {"name": "Execute Before", "type": "u128"},
                    {"name": "Calls", "type": "Call*"},
                ],
                "Call": [
                    {"name": "To", "type": "ContractAddress"},
                    {"name": "Selector", "type": "selector"},
                    {"name": "Calldata", "type": "felt*"},
                ],
            },
            "primaryType": "OutsideExecution",
            "domain": {
                "name": "Account.execute_from_outside",
                "version": "2",
                "chainId": str(chain_id),
                "revision": Revision.V1,
            },
            "message": {
                "Caller": outside_execution.caller,
                "Nonce": outside_execution.nonce,
                "Execute After": outside_execution.execute_after,
                "Execute Before": outside_execution.execute_before,
                "Calls": [
                    {
                        "To": call.to_addr,
                        "Selector": call.selector,
                        "Calldata": call.calldata,
                    }
                    for call in outside_execution.calls
                ],
            },
        }
    )



from dataclasses import dataclass
from typing import Any, Generator

from starknet_py.serialization._context import (
    Context,
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)


@dataclass
class NonZeroSerializer(CairoDataSerializer[Any, int]):
    """
    Serializer for NonZero type.
    Can serialize Cairo types which are non-zero.
    Deserializes data to int.
    """

    serializer: CairoDataSerializer

    def deserialize_with_context(self, context: DeserializationContext) -> int:
        deserialized_value = self.serializer.deserialize_with_context(context)
        self._ensure_valid_nonzero(deserialized_value, context)
        return deserialized_value

    def serialize_with_context(
        self,
        context: SerializationContext,
        value: Any,
    ) -> Generator[int, None, None]:
        self._ensure_valid_nonzero(value, context)
        return self.serializer.serialize_with_context(context, value)

    @staticmethod
    def _ensure_valid_nonzero(value: int, context: Context):
        context.ensure_valid_value(value != 0, "expected value to be non-zero")



import dataclasses

import pytest

from starknet_py.constants import ADDR_BOUND
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_models import (
    Call,
    DAMode,
    ResourceBoundsMapping,
    Transaction,
    TransactionType,
    TransactionV3,
)
from starknet_py.net.full_node_client import _create_broadcasted_txn, _to_storage_key
from starknet_py.net.http_client import RpcHttpClient, ServerError
from starknet_py.net.models.transaction import DeclareV3, DeployAccountV3, InvokeV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_wait_for_tx_negative_check_interval(client):
    with pytest.raises(
        ValueError, match="Argument check_interval has to be greater than 0."
    ):
        await client.wait_for_tx(tx_hash=0, check_interval=-1)


def test_cannot_instantiate_abstract_transaction_class():
    with pytest.raises(
        TypeError, match="Cannot instantiate abstract Transaction class."
    ):
        _ = Transaction(hash=0, signature=[0, 0], version=0)


def test_cannot_instantiate_abstract_transaction_v3_class():
    with pytest.raises(
        TypeError, match="Cannot instantiate abstract TransactionV3 class."
    ):
        _ = TransactionV3(
            hash=0,
            signature=[0, 0],
            version=0,
            paymaster_data=[],
            tip=0,
            nonce_data_availability_mode=DAMode.L1,
            fee_data_availability_mode=DAMode.L1,
            resource_bounds=ResourceBoundsMapping.init_with_zeros(),
        )


def test_handle_rpc_error_server_error():
    no_error_dict = {"not_an_error": "success"}

    with pytest.raises(ServerError, match="RPC request failed."):
        RpcHttpClient.handle_rpc_error(no_error_dict)


@pytest.mark.parametrize(
    "key, expected",
    [
        (0x0, "0x00"),
        (0x12345, "0x012345"),
        (0x10001, "0x010001"),
        (0xFFAA, "0x00ffaa"),
        (0xDE, "0x00de"),
        (
            ADDR_BOUND - 1,
            "0x07fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeff",
        ),
    ],
)
def test_get_rpc_storage_key(key, expected):
    assert _to_storage_key(key) == expected


@pytest.mark.parametrize(
    "key",
    [int(1e100), -1, 0x8FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF],
)
def test_get_rpc_storage_key_raises_on_non_representable_key(key):
    with pytest.raises(ValueError, match="cannot be represented"):
        _to_storage_key(key)


@pytest.mark.asyncio
async def test_broadcasted_txn_declare_v3(
    account, abi_types_compiled_contract_and_class_hash
):
    declare_v3 = await account.sign_declare_v3(
        compiled_contract=abi_types_compiled_contract_and_class_hash[0],
        compiled_class_hash=abi_types_compiled_contract_and_class_hash[1],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    brodcasted_txn = _create_broadcasted_txn(declare_v3)
    assert brodcasted_txn["type"] == TransactionType.DECLARE.name

    expected_keys = dataclasses.fields(DeclareV3)
    assert all(key.name in brodcasted_txn for key in expected_keys)


@pytest.mark.asyncio
async def test_broadcasted_txn_invoke_v3(account, hello_starknet_contract):
    invoke_tx = await account.sign_invoke_v3(
        calls=Call(
            hello_starknet_contract.address,
            get_selector_from_name("increaseBalance"),
            [10],
        ),
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    brodcasted_txn = _create_broadcasted_txn(invoke_tx)

    assert brodcasted_txn["type"] == TransactionType.INVOKE.name

    expected_keys = dataclasses.fields(InvokeV3)
    assert all(key.name in brodcasted_txn for key in expected_keys)


@pytest.mark.asyncio
async def test_broadcasted_txn_deploy_account_v3(account):
    class_hash = 0x1234
    salt = 0x123
    calldata = [1, 2, 3]
    signed_tx = await account.sign_deploy_account_v3(
        class_hash,
        salt,
        resource_bounds=MAX_RESOURCE_BOUNDS,
        constructor_calldata=calldata,
    )
    brodcasted_txn = _create_broadcasted_txn(signed_tx)
    assert brodcasted_txn["type"] == TransactionType.DEPLOY_ACCOUNT.name

    expected_keys = dataclasses.fields(DeployAccountV3)
    assert all(key.name in brodcasted_txn for key in expected_keys)



import json
from typing import cast

import pytest

from starknet_py.abi.v2 import Abi, AbiParser
from starknet_py.cairo.data_types import UintType
from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract


@pytest.mark.parametrize(
    "contract_name",
    [
        "AbiTypes",
        "Account",
        "ERC20",
        "Hello2",
        "HelloStarknet",
        "MinimalContract",
        "NewSyntaxTestContract",
        "TestContract",
        "TestEnum",
        "TestOption",
        "TokenBridge",
        "l1_l2",
    ],
)
def test_abi_parse(contract_name):
    abi = json.loads(
        load_contract(contract_name=contract_name, version=ContractVersion.V2)["sierra"]
    )["abi"]

    parser = AbiParser(abi)
    parsed_abi = parser.parse()

    assert isinstance(parsed_abi, Abi)


def test_bounded_int_parse_pre_2_8_0():
    abi_list = [
        {
            "type": "struct",
            "name": "core::circuit::u384",
            "members": [
                {
                    "name": "limb0",
                    "type": "core::internal::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb1",
                    "type": "core::internal::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb2",
                    "type": "core::internal::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb3",
                    "type": "core::internal::BoundedInt::<0, 79228162514264337593543950335>",
                },
            ],
        }
    ]

    parser = AbiParser(abi_list)
    parsed_abi = parser.parse()

    assert isinstance(parsed_abi, Abi)

    uint = cast(
        UintType, parsed_abi.defined_structures["core::circuit::u384"].types["limb0"]
    )
    assert uint.bits == 96


def test_bounded_int_parse_post_2_8_0():
    abi_list = [
        {
            "type": "struct",
            "name": "core::circuit::u384",
            "members": [
                {
                    "name": "limb0",
                    "type": "core::internal::bounded_int::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb1",
                    "type": "core::internal::bounded_int::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb2",
                    "type": "core::internal::bounded_int::BoundedInt::<0, 79228162514264337593543950335>",
                },
                {
                    "name": "limb3",
                    "type": "core::internal::bounded_int::BoundedInt::<0, 79228162514264337593543950335>",
                },
            ],
        }
    ]

    parser = AbiParser(abi_list)
    parsed_abi = parser.parse()

    assert isinstance(parsed_abi, Abi)

    uint = cast(
        UintType, parsed_abi.defined_structures["core::circuit::u384"].types["limb0"]
    )
    assert uint.bits == 96



from dataclasses import dataclass, field
from typing import List, Tuple

from starknet_py.hash.hash_method import HashMethod


@dataclass
class MerkleTree:
    """
    Dataclass representing a MerkleTree object.
    """

    leaves: List[int]
    hash_method: HashMethod
    root_hash: int = field(init=False)
    levels: List[List[int]] = field(init=False)

    def __post_init__(self):
        self.root_hash, self.levels = self._build()

    def _build(self) -> Tuple[int, List[List[int]]]:
        if not self.leaves:
            raise ValueError("Cannot build Merkle tree from an empty list of leaves.")

        if len(self.leaves) == 1:
            return self.leaves[0], [self.leaves]

        curr_level_nodes = self.leaves[:]
        levels: List[List[int]] = []

        while len(curr_level_nodes) > 1:
            if len(curr_level_nodes) != len(self.leaves):
                levels.append(curr_level_nodes[:])

            new_nodes = []
            for i in range(0, len(curr_level_nodes), 2):
                a, b = (
                    curr_level_nodes[i],
                    curr_level_nodes[i + 1] if i + 1 < len(curr_level_nodes) else 0,
                )
                new_nodes.append(self.hash_method.hash(*sorted([a, b])))

            curr_level_nodes = new_nodes
        levels = [self.leaves] + levels + [curr_level_nodes]
        return curr_level_nodes[0], levels



from marshmallow import fields, post_load
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from starknet_py.net.client_models import (
    DAMode,
    DeclareTransactionResponse,
    DeclareTransactionV0,
    DeclareTransactionV1,
    DeclareTransactionV2,
    DeclareTransactionV3,
    DeployAccountTransactionResponse,
    DeployAccountTransactionV1,
    DeployAccountTransactionV3,
    DeployTransaction,
    FeePayment,
    InvokeTransactionV0,
    InvokeTransactionV1,
    InvokeTransactionV3,
    L1HandlerTransaction,
    L2toL1Message,
    MessageStatus,
    ResourceBounds,
    ResourceBoundsMapping,
    SentTransactionResponse,
    TransactionReceipt,
    TransactionReceiptWithBlockInfo,
    TransactionStatusResponse,
    TransactionWithReceipt,
)
from starknet_py.net.schemas.common import (
    DAModeField,
    ExecutionStatusField,
    Felt,
    FinalityStatusField,
    NumberAsHex,
    PriceUnitField,
    StatusField,
    TransactionFinalityStatusField,
    TransactionTypeField,
    Uint64,
    Uint128,
)
from starknet_py.net.schemas.rpc.event import EventSchema
from starknet_py.net.schemas.rpc.general import ExecutionResourcesSchema
from starknet_py.net.schemas.utils import _extract_tx_version
from starknet_py.utils.schema import Schema


class L2toL1MessageSchema(Schema):
    l2_address = Felt(data_key="from_address", required=True)
    l1_address = Felt(data_key="to_address", required=True)
    payload = fields.List(Felt(), data_key="payload", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L2toL1Message:
        return L2toL1Message(**data)


class FeePaymentSchema(Schema):
    amount = Felt(data_key="amount", required=True)
    unit = PriceUnitField(data_key="unit", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> FeePayment:
        return FeePayment(**data)


class TransactionReceiptSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)
    execution_status = ExecutionStatusField(data_key="execution_status", required=True)
    finality_status = FinalityStatusField(data_key="finality_status", required=True)
    actual_fee = fields.Nested(FeePaymentSchema(), data_key="actual_fee", required=True)
    type = TransactionTypeField(data_key="type", required=True)
    contract_address = Felt(data_key="contract_address", load_default=None)
    revert_reason = fields.String(data_key="revert_reason", load_default=None)
    events = fields.List(
        fields.Nested(EventSchema()), data_key="events", load_default=[]
    )
    messages_sent = fields.List(
        fields.Nested(L2toL1MessageSchema()), data_key="messages_sent", load_default=[]
    )
    message_hash = NumberAsHex(data_key="message_hash", load_default=None)
    execution_resources = fields.Nested(
        ExecutionResourcesSchema(), data_key="execution_resources", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionReceipt:
        return TransactionReceipt(**data)


class TransactionReceiptWithBlockInfoSchema(TransactionReceiptSchema):
    block_hash = Felt(data_key="block_hash", load_default=None)
    block_number = fields.Integer(data_key="block_number", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionReceiptWithBlockInfo:
        return TransactionReceiptWithBlockInfo(**data)


class TransactionStatusResponseSchema(Schema):
    finality_status = StatusField(data_key="finality_status", required=True)
    execution_status = ExecutionStatusField(
        data_key="execution_status", load_default=None
    )
    failure_reason = fields.String(data_key="failure_reason", load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionStatusResponse:
        return TransactionStatusResponse(**data)


class ResourceBoundsSchema(Schema):
    max_amount = Uint64(data_key="max_amount", required=True)
    max_price_per_unit = Uint128(data_key="max_price_per_unit", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> ResourceBounds:
        return ResourceBounds(**data)


class ResourceBoundsMappingSchema(Schema):
    l1_gas = fields.Nested(ResourceBoundsSchema(), data_key="l1_gas", required=True)
    l2_gas = fields.Nested(ResourceBoundsSchema(), data_key="l2_gas", required=True)
    l1_data_gas = fields.Nested(
        ResourceBoundsSchema(), data_key="l1_data_gas", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ResourceBoundsMapping:
        return ResourceBoundsMapping(**data)


class TransactionSchema(Schema):
    hash = Felt(data_key="transaction_hash", load_default=None)
    signature = fields.List(Felt(), data_key="signature", load_default=[])
    version = Felt(data_key="version", required=True)


class DeprecatedTransactionSchema(TransactionSchema):
    max_fee = Felt(data_key="max_fee", required=True)


class TransactionV3Schema(TransactionSchema):
    tip = Uint64(data_key="tip", required=True)
    nonce_data_availability_mode = DAModeField(
        data_key="nonce_data_availability_mode", load_default=DAMode.L1
    )
    fee_data_availability_mode = DAModeField(
        data_key="fee_data_availability_mode", load_default=DAMode.L1
    )
    paymaster_data = fields.List(Felt(), data_key="paymaster_data", load_default=[])
    resource_bounds = fields.Nested(
        ResourceBoundsMappingSchema(), data_key="resource_bounds", required=True
    )


class InvokeTransactionV0Schema(DeprecatedTransactionSchema):
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    contract_address = Felt(data_key="contract_address", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)

    @post_load
    def make_transaction(self, data, **kwargs) -> InvokeTransactionV0:
        return InvokeTransactionV0(**data)


class InvokeTransactionV1Schema(DeprecatedTransactionSchema):
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    sender_address = Felt(data_key="sender_address", required=True)
    nonce = Felt(data_key="nonce", required=True)

    @post_load
    def make_transaction(self, data, **kwargs) -> InvokeTransactionV1:
        return InvokeTransactionV1(**data)


class InvokeTransactionV3Schema(TransactionV3Schema):
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    sender_address = Felt(data_key="sender_address", required=True)
    nonce = Felt(data_key="nonce", required=True)
    account_deployment_data = fields.List(
        Felt(), data_key="account_deployment_data", required=True
    )

    @post_load
    def make_transaction(self, data, **kwargs) -> InvokeTransactionV3:
        return InvokeTransactionV3(**data)


class DeclareTransactionV0Schema(DeprecatedTransactionSchema):
    sender_address = Felt(data_key="sender_address", required=True)
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionV0:
        return DeclareTransactionV0(**data)


class DeclareTransactionV1Schema(DeprecatedTransactionSchema):
    sender_address = Felt(data_key="sender_address", required=True)
    class_hash = Felt(data_key="class_hash", required=True)
    nonce = Felt(data_key="nonce", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionV1:
        return DeclareTransactionV1(**data)


class DeclareTransactionV2Schema(DeprecatedTransactionSchema):
    sender_address = Felt(data_key="sender_address", required=True)
    class_hash = Felt(data_key="class_hash", required=True)
    compiled_class_hash = Felt(data_key="compiled_class_hash", required=True)
    nonce = Felt(data_key="nonce", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionV2:
        return DeclareTransactionV2(**data)


class DeclareTransactionV3Schema(TransactionV3Schema):
    sender_address = Felt(data_key="sender_address", required=True)
    class_hash = Felt(data_key="class_hash", required=True)

    compiled_class_hash = Felt(data_key="compiled_class_hash", required=True)
    nonce = Felt(data_key="nonce", required=True)
    account_deployment_data = fields.List(
        Felt(), data_key="account_deployment_data", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionV3:
        return DeclareTransactionV3(**data)


class DeployTransactionSchema(TransactionSchema):
    contract_address_salt = Felt(data_key="contract_address_salt", required=True)
    constructor_calldata = fields.List(
        Felt(), data_key="constructor_calldata", required=True
    )
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployTransaction:
        return DeployTransaction(**data)


class DeployAccountTransactionV1Schema(DeprecatedTransactionSchema):
    nonce = Felt(data_key="nonce", required=True)
    contract_address_salt = Felt(data_key="contract_address_salt", required=True)
    constructor_calldata = fields.List(
        Felt(), data_key="constructor_calldata", required=True
    )
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployAccountTransactionV1:
        return DeployAccountTransactionV1(**data)


class DeployAccountTransactionV3Schema(TransactionV3Schema):
    nonce = Felt(data_key="nonce", required=True)
    contract_address_salt = Felt(data_key="contract_address_salt", required=True)
    constructor_calldata = fields.List(
        Felt(), data_key="constructor_calldata", required=True
    )
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployAccountTransactionV3:
        return DeployAccountTransactionV3(**data)


class DeclareTransactionSchema(OneOfSchema):
    type_schemas = {
        "0": DeclareTransactionV0Schema,
        "1": DeclareTransactionV1Schema,
        "2": DeclareTransactionV2Schema,
        "3": DeclareTransactionV3Schema,
    }

    def get_data_type(self, data):
        return _extract_tx_version(data.get("version"))


class InvokeTransactionSchema(OneOfSchema):
    type_schemas = {
        "0": InvokeTransactionV0Schema,
        "1": InvokeTransactionV1Schema,
        "3": InvokeTransactionV3Schema,
    }

    def get_obj_type(self, obj):
        return _extract_tx_version(obj.version)

    def get_data_type(self, data):
        return _extract_tx_version(data.get("version"))


class DeployAccountTransactionSchema(OneOfSchema):
    type_schemas = {
        "1": DeployAccountTransactionV1Schema,
        "3": DeployAccountTransactionV3Schema,
    }

    def get_obj_type(self, obj):
        return _extract_tx_version(obj.version)

    def get_data_type(self, data):
        return _extract_tx_version(data.get("version"))


class L1HandlerTransactionSchema(TransactionSchema):
    contract_address = Felt(data_key="contract_address", required=True)
    calldata = fields.List(Felt(), data_key="calldata", required=True)
    entry_point_selector = Felt(data_key="entry_point_selector", required=True)
    nonce = NumberAsHex(data_key="nonce", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> L1HandlerTransaction:
        return L1HandlerTransaction(**data)


class TypesOfTransactionsSchema(OneOfSchema):
    type_field = "type"
    type_schemas = {
        "INVOKE": InvokeTransactionSchema,
        "DECLARE": DeclareTransactionSchema,
        "DEPLOY": DeployTransactionSchema,
        "DEPLOY_ACCOUNT": DeployAccountTransactionSchema,
        "L1_HANDLER": L1HandlerTransactionSchema,
    }


class TransactionWithReceiptSchema(Schema):
    transaction = fields.Nested(TypesOfTransactionsSchema(), data_key="transaction")
    receipt = fields.Nested(TransactionReceiptSchema(), data_key="receipt")

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionWithReceipt:
        return TransactionWithReceipt(**data)


class SentTransactionSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SentTransactionResponse:
        return SentTransactionResponse(**data)


class DeclareTransactionResponseSchema(SentTransactionSchema):
    class_hash = Felt(data_key="class_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeclareTransactionResponse:
        return DeclareTransactionResponse(**data)


class DeployAccountTransactionResponseSchema(SentTransactionSchema):
    address = Felt(data_key="contract_address", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployAccountTransactionResponse:
        return DeployAccountTransactionResponse(**data)


class MessageStatusSchema(Schema):
    transaction_hash = NumberAsHex(data_key="transaction_hash", required=True)
    finality_status = TransactionFinalityStatusField(
        data_key="finality_status", required=True
    )
    execution_status = ExecutionStatusField(data_key="execution_status", required=True)
    failure_reason = fields.String(data_key="failure_reason", load_default=None)

    @post_load
    def make_dataclass(self, data, **kwargs) -> MessageStatus:
        return MessageStatus(**data)



import os
import sys

import pytest

from starknet_py.net.client_models import ResourceBoundsMapping

directory = os.path.dirname(__file__)


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_using_account(account, map_compiled_contract_and_class_hash_copy_2):
    # pylint: disable=import-outside-toplevel, duplicate-code, too-many-locals
    (compiled_contract, class_hash) = map_compiled_contract_and_class_hash_copy_2
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds

    # docs: end
    # docs: start
    l1_resource_bounds = ResourceBounds(
        max_amount=int(1e5), max_price_per_unit=int(1e13)
    )
    resource_bounds = ResourceBoundsMapping(
        l1_gas=l1_resource_bounds,
        l2_gas=ResourceBounds.init_with_zeros(),
        l1_data_gas=ResourceBounds.init_with_zeros(),
    )
    # Declare and deploy an example contract which implements a simple k-v store.
    # Contract.declare_v3 takes string containing a compiled contract (sierra) and
    # a class hash (casm_class_hash) or string containing a compiled contract (casm)
    declare_result = await Contract.declare_v3(
        account,
        compiled_contract=compiled_contract,
        compiled_class_hash=class_hash,
        resource_bounds=resource_bounds,
    )

    await declare_result.wait_for_acceptance()
    deploy_result = await declare_result.deploy_v3(
        resource_bounds=resource_bounds,
    )
    # Wait until deployment transaction is accepted
    await deploy_result.wait_for_acceptance()

    # Get deployed contract
    map_contract = deploy_result.deployed_contract
    k, v = 13, 4324
    # Adds a transaction to mutate the state of k-v store. The call goes through account proxy, because we've used
    # Account to create the contract object
    resource_bounds = ResourceBoundsMapping(
        l1_gas=l1_resource_bounds,
        l2_gas=ResourceBounds.init_with_zeros(),
        l1_data_gas=ResourceBounds.init_with_zeros(),
    )
    await (
        await map_contract.functions["put"].invoke_v3(
            k,
            v,
            resource_bounds=resource_bounds,
        )
    ).wait_for_acceptance()

    # Retrieves the value, which is equal to 4324 in this case
    (resp,) = await map_contract.functions["get"].call(k)

    # There is a possibility of invoking the multicall

    # Creates a list of prepared function calls
    calls = [
        map_contract.functions["put"].prepare_invoke_v3(key=10, value=20),
        map_contract.functions["put"].prepare_invoke_v3(key=30, value=40),
    ]

    # Executes only one transaction with prepared calls
    transaction_response = await account.execute_v3(
        calls=calls,
        resource_bounds=resource_bounds,
    )
    await account.client.wait_for_tx(transaction_response.transaction_hash)
    # docs: end

    assert resp == v



import re
from typing import Dict, Optional, Union, cast

from typing_extensions import get_args

from starknet_py.hash.utils import encode_uint, encode_uint_list
from starknet_py.net.client_models import Hash, L1HandlerTransaction, Tag
from starknet_py.net.models.transaction import AccountTransaction
from starknet_py.net.schemas.broadcasted_txn import BroadcastedTransactionSchema


def hash_to_felt(value: Hash) -> str:
    """
    Convert hash to hexadecimal string
    """
    if isinstance(value, str):
        return value

    return hex(value)


def is_block_identifier(value: Union[int, Hash, Tag]) -> bool:
    return isinstance(value, str) and value in get_args(Tag)


def encode_l1_message(tx: L1HandlerTransaction) -> bytes:
    from_address = tx.calldata[0]
    # Pop first element to have in calldata the actual payload
    tx.calldata.pop(0)

    return (
        encode_uint(from_address)
        + encode_uint(tx.contract_address)
        + encode_uint(tx.nonce)
        + encode_uint(tx.entry_point_selector)
        + encode_uint(len(tx.calldata))
        + encode_uint_list(tx.calldata)
    )


def _to_storage_key(key: int) -> str:
    """
    Convert a value to RPC storage key matching a ``^0x0[0-7]{1}[a-fA-F0-9]{0,62}$`` pattern.

    :param key: The key to convert.
    :return: RPC storage key representation of the key.
    """

    hashed_key = hex(key)[2:]

    if hashed_key[0] not in ("0", "1", "2", "3", "4", "5", "6", "7"):
        hashed_key = "0" + hashed_key

    hashed_key = "0x0" + hashed_key

    if not re.match(r"^0x0[0-7]{1}[a-fA-F0-9]{0,62}$", hashed_key):
        raise ValueError(f"Value {key} cannot be represented as RPC storage key.")

    return hashed_key


def _to_rpc_felt(value: Hash) -> str:
    """
    Convert the value to RPC felt matching a ``^0x(0|[a-fA-F1-9]{1}[a-fA-F0-9]{0,62})$`` pattern.

    :param value: The value to convert.
    :return: RPC felt representation of the value.
    """
    if isinstance(value, str):
        value = int(value, 16)

    rpc_felt = hex(value)
    assert re.match(r"^0x(0|[a-fA-F1-9]{1}[a-fA-F0-9]{0,62})$", rpc_felt)
    return rpc_felt


def _is_valid_eth_address(address: str) -> bool:
    """
    A function checking if an address matches Ethereum address regex. Note that it doesn't validate any checksums etc.
    """
    return bool(re.fullmatch("^0x[a-fA-F0-9]{40}$", address))


def _create_broadcasted_txn(transaction: AccountTransaction) -> dict:
    return cast(
        Dict,
        BroadcastedTransactionSchema().dump(obj=transaction),
    )


def get_block_identifier(
    block_hash: Optional[Union[Hash, Tag]] = None,
    block_number: Optional[Union[int, Tag]] = None,
    default_tag: Optional[Tag] = "pre_confirmed",
    allow_pre_confirmed: bool = True,
) -> dict:
    block_id = {
        "block_id": _get_raw_block_identifier(block_hash, block_number, default_tag)
    }
    if not allow_pre_confirmed and block_id == {"block_id": "pre_confirmed"}:
        raise ValueError("Value pre_confirmed is not valid for this method.")
    return block_id


def _get_raw_block_identifier(
    block_hash: Optional[Union[Hash, Tag]] = None,
    block_number: Optional[Union[int, Tag]] = None,
    default_tag: Optional[Tag] = "pre_confirmed",
) -> Union[dict, Hash, Tag, None]:
    if block_hash is not None and block_number is not None:
        raise ValueError(
            "Arguments block_hash and block_number are mutually exclusive."
        )

    block_tags = get_args(Tag)

    if block_hash in block_tags or block_number in block_tags:
        return block_hash or block_number

    if block_hash is not None:
        return {"block_hash": _to_rpc_felt(block_hash)}

    if block_number is not None:
        return {"block_number": block_number}

    return default_tag



from dataclasses import dataclass
from typing import Generator

from starknet_py.serialization._context import (
    Context,
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)


@dataclass
class BoolSerializer(CairoDataSerializer[bool, int]):
    """
    Serializer for boolean.
    """

    def deserialize_with_context(self, context: DeserializationContext) -> bool:
        [val] = context.reader.read(1)
        self._ensure_bool(context, val)
        return bool(val)

    def serialize_with_context(
        self, context: SerializationContext, value: bool
    ) -> Generator[int, None, None]:
        context.ensure_valid_type(value, isinstance(value, bool), "bool")
        self._ensure_bool(context, value)
        yield int(value)

    @staticmethod
    def _ensure_bool(context: Context, value: int):
        context.ensure_valid_value(
            value in [0, 1],
            f"invalid value '{value}' - must be in [0, 2) range",
        )



from typing import Any, List, Optional, Tuple

import lark
from lark import Token, Transformer

from starknet_py.cairo.data_types import (
    ArrayType,
    BoolType,
    CairoType,
    FeltType,
    FixedSizeArrayType,
    OptionType,
    TupleType,
    TypeIdentifier,
    UintType,
    UnitType,
)

ABI_EBNF = """
    IDENTIFIER: /[a-zA-Z_][a-zA-Z_0-9]*/
    
    type: type_unit
        | type_bool
        | type_felt
        | type_uint
        | type_contract_address
        | type_class_hash
        | type_storage_address
        | type_option
        | type_array
        | type_fixed_size_array
        | type_span
        | tuple
        | type_identifier
    
    
    type_unit: "()"
    type_felt: "core::felt252"
    type_bool: "core::bool"
    type_uint: "core::integer::u" INT
    type_contract_address: "core::starknet::contract_address::ContractAddress"
    type_class_hash: "core::starknet::class_hash::ClassHash"
    type_storage_address: "core::starknet::storage_access::StorageAddress"
    type_option: "core::option::Option::<" (type | type_identifier) ">"
    type_array: "core::array::Array::<" (type | type_identifier) ">"
    type_fixed_size_array: "[" (type | type_identifier) ";" INT "]"
    type_span: "core::array::Span::<" (type | type_identifier) ">"
    
    tuple: "(" type? ("," type?)* ")"
    
    type_identifier: (IDENTIFIER | "::")+ ("<" (type | ",")+ ">")?
    
    
    %import common.INT
    %import common.WS
    %ignore WS
"""


class ParserTransformer(Transformer):
    """
    Transforms the lark tree into CairoTypes.
    """

    def __init__(self, type_identifiers: Optional[dict] = None) -> None:
        if type_identifiers is None:
            type_identifiers = {}
        self.type_identifiers = type_identifiers
        super(Transformer, self).__init__()

    # pylint: disable=no-self-use

    def __default__(self, data: str, children, meta):
        raise TypeError(f"Unable to parse tree node of type {data}.")

    def type(self, value: List[Optional[CairoType]]) -> Optional[CairoType]:
        """
        Tokens are read bottom-up, so here all of them are parsed and should be just returned.
        `Optional` is added in case of the unit type.
        """
        assert len(value) == 1
        return value[0]

    def type_felt(self, _value: List[Any]) -> FeltType:
        """
        Felt does not contain any additional arguments, so `_value` is just an empty list.
        """
        return FeltType()

    def type_bool(self, _value: List[Any]) -> BoolType:
        """
        Bool does not contain any additional arguments, so `_value` is just an empty list.
        """
        return BoolType()

    def type_uint(self, value: List[Token]) -> UintType:
        """
        Uint type contains information about its size. It is present in the value[0].
        """
        return UintType(int(value[0]))

    def type_unit(self, _value: List[Any]) -> UnitType:
        """
        `()` type.
        """
        return UnitType()

    def type_option(self, value: List[CairoType]) -> OptionType:
        """
        Option includes an information about which type it eventually represents.
        `Optional` is added in case of the unit type.
        """
        return OptionType(value[0])

    def type_array(self, value: List[CairoType]) -> ArrayType:
        """
        Array contains values of type under `value[0]`.
        """
        return ArrayType(value[0])

    def type_fixed_size_array(
        self, value: Tuple[CairoType, Token]
    ) -> FixedSizeArrayType:
        """
        Fixed-size array contains values of type under `value[0]`.
        """
        cairo_type, size_token = value
        size = int(size_token)
        return FixedSizeArrayType(cairo_type, size)

    def type_span(self, value: List[CairoType]) -> ArrayType:
        """
        Span contains values of type under `value[0]`.
        """
        return ArrayType(value[0])

    def type_identifier(self, tokens: List[Token]) -> TypeIdentifier:
        """
        Structs and enums are defined as follows: (IDENTIFIER | "::")+ [some not important info]
        where IDENTIFIER is a string.

        Tokens would contain strings and types (if it is present).
        We are interested only in the strings because a structure (or enum) name can be built from them.
        """
        name = "::".join(token for token in tokens if isinstance(token, str))
        if name in self.type_identifiers:
            return self.type_identifiers[name]
        return TypeIdentifier(name)

    def type_contract_address(self, _value: List[Any]) -> FeltType:
        """
        ContractAddress is represented by the felt252.
        """
        return FeltType()

    def type_class_hash(self, _value: List[Any]) -> FeltType:
        """
        ClassHash is represented by the felt252.
        """
        return FeltType()

    def type_storage_address(self, _value: List[Any]) -> FeltType:
        """
        StorageAddress is represented by the felt252.
        """
        return FeltType()

    def tuple(self, types: List[CairoType]) -> TupleType:
        """
        Tuple contains values defined in the `types` argument.
        """
        return TupleType(types)


def parse(
    code: str,
    type_identifiers,
) -> CairoType:
    """
    Parse the given string and return a CairoType.
    """
    grammar_parser = lark.Lark(
        grammar=ABI_EBNF,
        start="type",
        parser="earley",
    )
    parsed = grammar_parser.parse(code)

    parser_transformer = ParserTransformer(type_identifiers)
    cairo_type = parser_transformer.transform(parsed)

    return cairo_type



import sys

import pytest

from starknet_py.cairo.felt import decode_shortstring, encode_shortstring
from starknet_py.contract import Contract
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.contracts_v1 import deploy_v3_contract

# TODO (#1219): investigate why some of these tests fails for contracts_compiled_v1


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_general_v3_interaction(account, erc20_class_hash: int):
    calldata = {
        "name_": encode_shortstring("erc20_basic"),
        "symbol_": encode_shortstring("ERC20B"),
        "decimals_": 10,
        "initial_supply": 12345,
        "recipient": account.address,
    }
    erc20 = await deploy_v3_contract(
        account=account,
        contract_name="ERC20",
        class_hash=erc20_class_hash,
        calldata=calldata,
    )

    (name,) = await erc20.functions["get_name"].call()
    decoded_name = decode_shortstring(name)
    (decimals,) = await erc20.functions["get_decimals"].call()
    (supply,) = await erc20.functions["get_total_supply"].call()
    (account_balance,) = await erc20.functions["balance_of"].call(
        account=account.address
    )

    transfer_amount = 10
    await (
        await erc20.functions["transfer"].invoke_v3(
            recipient=0x11, amount=transfer_amount, resource_bounds=MAX_RESOURCE_BOUNDS
        )
    ).wait_for_acceptance()

    (after_transfer_balance,) = await erc20.functions["balance_of"].call(
        account=account.address
    )

    assert decoded_name == "erc20_basic"
    assert decimals == calldata["decimals_"]
    assert supply == calldata["initial_supply"]
    assert account_balance == calldata["initial_supply"]
    assert after_transfer_balance == calldata["initial_supply"] - transfer_amount


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_serializing_struct(account, token_bridge_class_hash: int):
    bridge = await deploy_v3_contract(
        account=account,
        contract_name="TokenBridge",
        class_hash=token_bridge_class_hash,
        calldata={"governor_address": account.address},
    )

    await (
        await bridge.functions["set_l1_bridge"].invoke_v3(
            l1_bridge_address={"address": 0x11}, resource_bounds=MAX_RESOURCE_BOUNDS
        )
    ).wait_for_acceptance()


@pytest.mark.asyncio
async def test_serializing_option(account, test_option_class_hash: int):
    test_option = await deploy_v3_contract(
        account=account,
        contract_name="TestOption",
        class_hash=test_option_class_hash,
    )

    (received_option,) = await test_option.functions["get_option_struct"].call()

    assert dict(received_option) == {
        "first_field": 1,
        "second_field": 2,
        "third_field": None,
        "fourth_field": 4,
    }

    option_struct = {
        "first_field": 1,
        "second_field": 2**128 + 1,
        "third_field": None,
        "fourth_field": 4,
    }

    (received_option,) = await test_option.functions[
        "receive_and_send_option_struct"
    ].call(option_struct=option_struct)

    assert dict(received_option) == option_struct

    (received_option,) = await test_option.functions["get_empty_option"].call()

    assert received_option is None


@pytest.mark.asyncio
async def test_serializing_enum(account, test_enum_class_hash: int):
    test_enum = await deploy_v3_contract(
        account=account,
        contract_name="TestEnum",
        class_hash=test_enum_class_hash,
    )

    (received_enum,) = await test_enum.functions["get_enum"].call()

    assert received_enum.variant == "a"
    assert received_enum.value == 100

    (received_enum,) = await test_enum.functions["get_enum_without_value"].call()

    assert received_enum.variant == "c"
    assert received_enum.value is None

    variant_name = "b"
    value = 200
    (received_enum,) = await test_enum.functions["receive_and_send_enum"].call(
        my_enum={variant_name: value}
    )

    assert received_enum.variant == variant_name
    assert received_enum.value == value

    variant_name = "c"
    value = None
    (received_enum,) = await test_enum.functions["receive_and_send_enum"].call(
        my_enum={variant_name: value}
    )

    assert received_enum.variant == variant_name
    assert received_enum.value == value


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_from_address_on_v1_contract(account, erc20_class_hash: int):
    calldata = {
        "name_": encode_shortstring("erc20_basic"),
        "symbol_": encode_shortstring("ERC20B"),
        "decimals_": 10,
        "initial_supply": 12345,
        "recipient": account.address,
    }
    erc20 = await deploy_v3_contract(
        account=account,
        contract_name="ERC20",
        class_hash=erc20_class_hash,
        calldata=calldata,
    )

    erc20_from_address = await Contract.from_address(erc20.address, provider=account)

    assert erc20_from_address.address == erc20.address
    assert erc20_from_address.account == erc20.account
    assert erc20_from_address.functions.keys() == erc20.functions.keys()
    assert erc20_from_address.data == erc20.data


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_invoke_contract_with_bytearray(string_contract):
    (initial_string,) = await string_contract.functions["get_string"].call()
    assert initial_string == "Hello"

    value_to_set = """
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
    """

    await string_contract.functions["set_string"].invoke_v3(
        new_string=value_to_set, auto_estimate=True
    )
    (new_string,) = await string_contract.functions["get_string"].call()
    assert new_string == value_to_set



from typing import List

import pytest
from poseidon_py.poseidon_hash import poseidon_hash

from starknet_py.hash.hash_method import HashMethod
from starknet_py.hash.utils import pedersen_hash
from starknet_py.utils.merkle_tree import MerkleTree


@pytest.mark.parametrize(
    "leaves, hash_method, expected_root_hash",
    [
        (
            ["0x12", "0xa"],
            HashMethod.PEDERSEN,
            "0x586699e3ba6f118227e094ad423313a2d51871507dcbc23116f11cdd79d80f2",
        ),
        (
            ["0x12", "0xa"],
            HashMethod.POSEIDON,
            "0x6257f1f60f7c9fd49e2718c8ad19cd8dce6b1ba4b553b2123113f22b1e9c379",
        ),
        (
            [
                "0x5bb9440e27889a364bcb678b1f679ecd1347acdedcbf36e83494f857cc58026",
                "0x3",
            ],
            HashMethod.PEDERSEN,
            "0x551b4adb6c35d49c686a00b9192da9332b18c9b262507cad0ece37f3b6918d2",
        ),
        (
            [
                "0x5bb9440e27889a364bcb678b1f679ecd1347acdedcbf36e83494f857cc58026",
                "0x3",
            ],
            HashMethod.POSEIDON,
            "0xc118a3963c12777b0717d1dc89baa8b3ceed84dfd713a6bd1354676f03f021",
        ),
    ],
)
def test_calculate_hash(
    leaves: List[str], hash_method: HashMethod, expected_root_hash: str
):
    if hash_method == HashMethod.PEDERSEN:
        apply_hash = pedersen_hash
    elif hash_method == HashMethod.POSEIDON:
        apply_hash = poseidon_hash
    else:
        raise ValueError(f"Unsupported hash method: {hash_method}.")

    a, b = int(leaves[0], 16), int(leaves[1], 16)
    merkle_hash = hash_method.hash(*sorted([b, a]))
    raw_hash = apply_hash(*sorted([b, a]))

    assert raw_hash == merkle_hash
    assert int(expected_root_hash, 16) == merkle_hash


@pytest.mark.parametrize(
    "hash_method",
    [
        HashMethod.PEDERSEN,
        HashMethod.POSEIDON,
    ],
)
def test_build_from_0_elements(hash_method: HashMethod):
    with pytest.raises(
        ValueError, match="Cannot build Merkle tree from an empty list of leaves."
    ):
        MerkleTree([], hash_method)


@pytest.mark.parametrize(
    "leaves, hash_method, expected_root_hash, expected_levels_count",
    [
        (["0x1"], HashMethod.PEDERSEN, "0x1", 1),
        (["0x1"], HashMethod.POSEIDON, "0x1", 1),
        (
            ["0x1", "0x2"],
            HashMethod.PEDERSEN,
            "0x5bb9440e27889a364bcb678b1f679ecd1347acdedcbf36e83494f857cc58026",
            2,
        ),
        (
            ["0x1", "0x2"],
            HashMethod.POSEIDON,
            "0x5d44a3decb2b2e0cc71071f7b802f45dd792d064f0fc7316c46514f70f9891a",
            2,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4"],
            HashMethod.PEDERSEN,
            "0x38118a340bbba28e678413cd3b07a9436a5e60fd6a7cbda7db958a6d501e274",
            3,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4"],
            HashMethod.POSEIDON,
            "0xa4d02f1e82fc554b062b754d3a4995e0ed8fc7e5016a7ca2894a451a4bae64",
            3,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4", "0x5", "0x6"],
            HashMethod.PEDERSEN,
            "0x329d5b51e352537e8424bfd85b34d0f30b77d213e9b09e2976e6f6374ecb59",
            4,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4", "0x5", "0x6"],
            HashMethod.POSEIDON,
            "0x34d525f018d8d6b3e492b1c9cda9bbdc3bc7834b408a30a417186c698c34766",
            4,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4", "0x5", "0x6", "0x7"],
            HashMethod.PEDERSEN,
            "0x7f748c75e5bdb7ae28013f076b8ab650c4e01d3530c6e5ab665f9f1accbe7d4",
            4,
        ),
        (
            ["0x1", "0x2", "0x3", "0x4", "0x5", "0x6", "0x7"],
            HashMethod.POSEIDON,
            "0x3308a3c50c25883753f82b21f14c644ec375b88ea5b0f83d1e6afe74d0ed790",
            4,
        ),
    ],
)
def test_build_from_elements(
    leaves: List[str],
    hash_method: HashMethod,
    expected_root_hash: str,
    expected_levels_count: int,
):
    tree = MerkleTree([int(leaf, 16) for leaf in leaves], hash_method)

    assert tree.root_hash is not None
    assert tree.levels is not None
    assert tree.root_hash == int(expected_root_hash, 16)
    assert len(tree.levels) == expected_levels_count



import warnings
from typing import Literal, Union, cast

from marshmallow import EXCLUDE, ValidationError

from starknet_py.net.client_models import (
    DeprecatedCompiledContract,
    DeprecatedContractClass,
    SierraCompiledContract,
)
from starknet_py.net.executable_models import CasmClass
from starknet_py.net.schemas.rpc.contract import (
    CasmClassSchema,
    ContractClassSchema,
    DeprecatedCompiledContractSchema,
    SierraCompiledContractSchema,
)


def create_compiled_contract(
    compiled_contract: str,
) -> DeprecatedCompiledContract:
    """
    Creates CompiledContract instance.

    :param compiled_contract: compiled contract string.
    :return: CompiledContract instance.
    """

    return cast(
        DeprecatedCompiledContract,
        DeprecatedCompiledContractSchema().loads(compiled_contract),
    )


def create_sierra_compiled_contract(compiled_contract: str) -> SierraCompiledContract:
    """
    Creates SierraCompiledContract instance.

    :param compiled_contract: compiled contract string.
    :return: SierraCompiledContract instance.
    """
    return cast(
        SierraCompiledContract,
        SierraCompiledContractSchema().loads(compiled_contract, unknown=EXCLUDE),
    )


def create_contract_class(
    compiled_contract: str,
) -> DeprecatedContractClass:
    """
    Creates ContractClass from already compiled contract.

     .. deprecated:: 0.15.0
        Function create_contract_class is deprecated and will be removed in the future.
        Use :meth:`create_compiled_contract` instead.

    :return: a ContractClass.
    """
    warnings.warn(
        "Function create_contract_class is deprecated and will be removed in the future. "
        "Consider using create_compiled_contract instead.",
        category=DeprecationWarning,
    )
    return cast(DeprecatedContractClass, ContractClassSchema().loads(compiled_contract))


def create_casm_class(compiled_contract: str) -> CasmClass:
    """
    Creates CasmClass instance.

    :param compiled_contract: contract compiled using starknet-sierra-compile.
    :return: CasmClass instance.
    """
    try:
        return cast(CasmClass, CasmClassSchema().loads(compiled_contract))
    except ValidationError as err:
        if err.messages == {"pythonic_hints": ["Missing data for required field."]}:
            raise ValueError(
                "Field pythonic_hints is missing from compiled_contract. "
                "Make sure to use starknet-sierra-compile with --add-pythonic-hints flag."
            ) from err
        raise err


def int_from_hex(number: Union[str, int]) -> int:
    return number if isinstance(number, int) else int(number, 16)


def int_from_bytes(
    value: bytes,
    byte_order: Literal["big", "little"] = "big",
    signed: bool = False,
) -> int:
    """
    Converts the given bytes object (parsed according to the given byte order) to an integer.
    """
    return int.from_bytes(value, byteorder=byte_order, signed=signed)



from dataclasses import dataclass
from enum import Enum
from typing import List, Literal, Optional, Tuple, Union

from starknet_py.net.client_models import CasmClassEntryPointsByType


class AssertCurrentAccessIndicesIsEmpty(Enum):
    ASSERT_CURRENT_ACCESS_INDICES_IS_EMPTY = "AssertCurrentAccessIndicesIsEmpty"


class AssertAllKeysUsed(Enum):
    ASSERT_ALL_KEYS_USED = "AssertAllKeysUsed"


class AssertLeAssertThirdArcExcluded(Enum):
    ASSERT_LE_ASSERT_THIRD_ARC_EXCLUDED = "AssertLeAssertThirdArcExcluded"


@dataclass
class CellRef:
    register: Literal["AP", "FP"]
    offset: int


@dataclass
class AssertAllAccessesUsedInner:
    n_used_accesses: CellRef


@dataclass
class AssertAllAccessesUsed:
    assert_all_accesses_used: AssertAllAccessesUsedInner


@dataclass
class Deref:
    deref: CellRef


@dataclass
class DoubleDeref:
    double_deref: Tuple[CellRef, int]


@dataclass
class Immediate:
    immediate: int


@dataclass
class BinOpInner:
    op: Literal["Add", "Mul"]
    a: CellRef
    b: Union[Deref, Immediate]


@dataclass
class BinOp:
    bin_op: BinOpInner


ResOperand = Union[Deref, DoubleDeref, Immediate, BinOp]


@dataclass
class AssertLtAssertValidInputInner:
    a: ResOperand
    b: ResOperand


@dataclass
class AssertLtAssertValidInput:
    assert_lt_assert_valid_input: AssertLtAssertValidInputInner


@dataclass
class Felt252DictReadInner:
    dict_ptr: ResOperand
    key: ResOperand
    value_dst: CellRef


@dataclass
class Felt252DictRead:
    felt252_dict_read: Felt252DictReadInner


@dataclass
class Felt252DictWriteInner:
    dict_ptr: ResOperand
    key: ResOperand
    value: ResOperand


@dataclass
class Felt252DictWrite:
    felt252_dict_write: Felt252DictWriteInner


@dataclass
class AllocSegmentInner:
    dst: CellRef


@dataclass
class AllocSegment:
    alloc_segment: AllocSegmentInner


@dataclass
class TestLessThanInner:
    lhs: ResOperand
    rhs: ResOperand
    dst: CellRef


@dataclass
class TestLessThan:
    test_less_than: TestLessThanInner


@dataclass
class TestLessThanOrEqualInner(TestLessThanInner):
    pass


@dataclass
class TestLessThanOrEqual:
    test_less_than_or_equal: TestLessThanOrEqualInner


@dataclass
class TestLessThanOrEqualAddressInner(TestLessThanInner):
    pass


@dataclass
class TestLessThanOrEqualAddress:
    test_less_than_or_equal_address: TestLessThanOrEqualAddressInner


@dataclass
class WideMul128Inner:
    lhs: ResOperand
    rhs: ResOperand
    high: CellRef
    low: CellRef


@dataclass
class WideMul128:
    wide_mul128: WideMul128Inner


@dataclass
class DivModInner:
    lhs: ResOperand
    rhs: ResOperand
    quotient: CellRef
    remainder: CellRef


@dataclass
class DivMod:
    div_mod: DivModInner


@dataclass
class Uint256DivModInner:
    # pylint: disable=too-many-instance-attributes
    dividend0: ResOperand
    dividend1: ResOperand
    divisor0: ResOperand
    divisor1: ResOperand
    quotient0: CellRef
    quotient1: CellRef
    remainder0: CellRef
    remainder1: CellRef


@dataclass
class Uint256DivMod:
    uint256_div_mod: Uint256DivModInner


@dataclass
class Uint512DivModByUint256Inner:
    # pylint: disable=too-many-instance-attributes
    dividend0: ResOperand
    dividend1: ResOperand
    dividend2: ResOperand
    dividend3: ResOperand
    divisor0: ResOperand
    divisor1: ResOperand
    quotient0: CellRef
    quotient1: CellRef
    quotient2: CellRef
    quotient3: CellRef
    remainder0: CellRef
    remainder1: CellRef


@dataclass
class Uint512DivModByUint256:
    uint512_div_mod_by_uint256: Uint512DivModByUint256Inner


@dataclass
class SquareRootInner:
    value: ResOperand
    dst: CellRef


@dataclass
class SquareRoot:
    square_root: SquareRootInner


@dataclass
class Uint256SquareRootInner:
    value_low: ResOperand
    value_high: ResOperand
    sqrt0: CellRef
    sqrt1: CellRef
    remainder_low: CellRef
    remainder_high: CellRef
    sqrt_mul_2_minus_remainder_ge_u128: CellRef


@dataclass
class Uint256SquareRoot:
    uint256_square_root: Uint256SquareRootInner


@dataclass
class LinearSplitInner:
    value: ResOperand
    scalar: ResOperand
    max_x: ResOperand
    x: CellRef
    y: CellRef


@dataclass
class LinearSplit:
    linear_split: LinearSplitInner


@dataclass
class AllocFelt252DictInner:
    segment_arena_ptr: ResOperand


@dataclass
class AllocFelt252Dict:
    alloc_felt252_dict: AllocFelt252DictInner


@dataclass
class Felt252DictEntryInitInner:
    dict_ptr: ResOperand
    key: ResOperand


@dataclass
class Felt252DictEntryInit:
    felt252_dict_entry_init: Felt252DictEntryInitInner


@dataclass
class Felt252DictEntryUpdateInner:
    dict_ptr: ResOperand
    value: ResOperand


@dataclass
class Felt252DictEntryUpdate:
    felt252_dict_entry_update: Felt252DictEntryUpdateInner


@dataclass
class GetSegmentArenaIndexInner:
    dict_end_ptr: ResOperand
    dict_index: CellRef


@dataclass
class GetSegmentArenaIndex:
    get_segment_arena_index: GetSegmentArenaIndexInner


@dataclass
class InitSquashDataInner:
    dict_accesses: ResOperand
    ptr_diff: ResOperand
    n_accesses: ResOperand
    big_keys: CellRef
    first_key: CellRef


@dataclass
class InitSquashData:
    init_squash_data: InitSquashDataInner


@dataclass
class GetCurrentAccessIndexInner:
    range_check_ptr: ResOperand


@dataclass
class GetCurrentAccessIndex:
    get_current_access_index: GetCurrentAccessIndexInner


@dataclass
class ShouldSkipSquashLoopInner:
    should_skip_loop: CellRef


@dataclass
class ShouldSkipSquashLoop:
    should_skip_squash_loop: ShouldSkipSquashLoopInner


@dataclass
class GetCurrentAccessDeltaInner:
    index_delta_minus1: CellRef


@dataclass
class GetCurrentAccessDelta:
    get_current_access_delta: GetCurrentAccessDeltaInner


@dataclass
class ShouldContinueSquashLoopInner:
    should_continue: CellRef


@dataclass
class ShouldContinueSquashLoop:
    should_continue_squash_loop: ShouldContinueSquashLoopInner


@dataclass
class GetNextDictKeyInner:
    next_key: CellRef


@dataclass
class GetNextDictKey:
    get_next_dict_key: GetNextDictKeyInner


@dataclass
class AssertLeFindSmallArcsInner:
    range_check_ptr: ResOperand
    a: ResOperand
    b: ResOperand


@dataclass
class AssertLeFindSmallArcs:
    assert_le_find_small_arcs: AssertLeFindSmallArcsInner


@dataclass
class AssertLeIsFirstArcExcludedInner:
    skip_exclude_a_flag: CellRef


@dataclass
class AssertLeIsFirstArcExcluded:
    assert_le_is_first_arc_excluded: AssertLeIsFirstArcExcludedInner


@dataclass
class AssertLeIsSecondArcExcludedInner:
    skip_exclude_b_minus_a: CellRef


@dataclass
class AssertLeIsSecondArcExcluded:
    assert_le_is_second_arc_excluded: AssertLeIsSecondArcExcludedInner


@dataclass
class RandomEcPointInner:
    x: CellRef
    y: CellRef


@dataclass
class RandomEcPoint:
    random_ec_point: RandomEcPointInner


@dataclass
class FieldSqrtInner:
    val: ResOperand
    sqrt: CellRef


@dataclass
class FieldSqrt:
    field_sqrt: FieldSqrtInner


@dataclass
class DebugPrintInner:
    start: ResOperand
    end: ResOperand


@dataclass
class DebugPrint:
    debug_print: DebugPrintInner


@dataclass
class AllocConstantSizeInner:
    size: ResOperand
    dst: CellRef


@dataclass
class AllocConstantSize:
    alloc_constant_size: AllocConstantSizeInner


@dataclass
class U256InvModNInner:
    # pylint: disable=too-many-instance-attributes
    b0: ResOperand
    b1: ResOperand
    n0: ResOperand
    n1: ResOperand
    g0_or_no_inv: CellRef
    g1_option: CellRef
    s_or_r0: CellRef
    s_or_r1: CellRef
    t_or_k0: CellRef
    t_or_k1: CellRef


@dataclass
class U256InvModN:
    u256_inv_mod_n: U256InvModNInner


@dataclass
class EvalCircuitInner:
    n_add_mods: ResOperand
    add_mod_builtin: ResOperand
    n_mul_mods: ResOperand
    mul_mod_builtin: ResOperand


@dataclass
class EvalCircuit:
    eval_circuit: EvalCircuitInner


@dataclass
class SystemCallInner:
    system: ResOperand


@dataclass
class SystemCall:
    system_call: SystemCallInner


@dataclass
class CheatcodeInner:
    selector: int
    input_start: ResOperand
    input_end: ResOperand
    output_start: CellRef
    output_end: CellRef


@dataclass
class Cheatcode:
    cheatcode: CheatcodeInner


Hint = Union[
    AssertCurrentAccessIndicesIsEmpty,
    AssertAllKeysUsed,
    AssertLeAssertThirdArcExcluded,
    AssertAllAccessesUsed,
    AssertLtAssertValidInput,
    Felt252DictRead,
    Felt252DictWrite,
    AllocSegment,
    TestLessThan,
    TestLessThanOrEqual,
    TestLessThanOrEqualAddress,
    WideMul128,
    DivMod,
    Uint256DivMod,
    Uint512DivModByUint256,
    SquareRoot,
    Uint256SquareRoot,
    LinearSplit,
    AllocFelt252Dict,
    Felt252DictEntryInit,
    Felt252DictEntryUpdate,
    GetSegmentArenaIndex,
    InitSquashData,
    GetCurrentAccessIndex,
    ShouldSkipSquashLoop,
    GetCurrentAccessDelta,
    ShouldContinueSquashLoop,
    GetNextDictKey,
    AssertLeFindSmallArcs,
    AssertLeIsFirstArcExcluded,
    AssertLeIsSecondArcExcluded,
    RandomEcPoint,
    FieldSqrt,
    DebugPrint,
    AllocConstantSize,
    U256InvModN,
    EvalCircuit,
    SystemCall,
    Cheatcode,
]


@dataclass
class CasmClass:
    """
    Dataclass representing class compiled to Cairo assembly.
    """

    prime: int
    bytecode: List[int]
    hints: List[Tuple[int, List[Hint]]]
    compiler_version: str
    entry_points_by_type: CasmClassEntryPointsByType
    bytecode_segment_lengths: Optional[List[int]]



from dataclasses import dataclass
from typing import Generator

from starknet_py.constants import FIELD_PRIME
from starknet_py.serialization._context import (
    Context,
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)


@dataclass
class IntSerializer(CairoDataSerializer[int, int]):
    """
    Serializer of int. In Cairo there are few ints (i8, ..., i64 and i128).
    Can serialize an int.
    Deserializes data to an int.
    """

    bits: int

    def deserialize_with_context(self, context: DeserializationContext) -> int:
        (raw,) = context.reader.read(1)

        signed_threshold = 1 << (self.bits - 1)
        deserialized_val = raw if raw < signed_threshold else raw - FIELD_PRIME
        with context.push_entity("int" + str(self.bits)):
            self._ensure_valid_int(
                deserialized_val,
                context,
                self.bits,
            )

            return deserialized_val

    def serialize_with_context(
        self, context: SerializationContext, value: int
    ) -> Generator[int, None, None]:
        context.ensure_valid_type(value, isinstance(value, int), "int")
        yield from self._serialize_from_int(value, context, self.bits)

    @staticmethod
    def _serialize_from_int(
        value: int, context: SerializationContext, bits: int
    ) -> Generator[int, None, None]:
        IntSerializer._ensure_valid_int(value, context, bits)

        unsigned = value % FIELD_PRIME

        yield unsigned

    @staticmethod
    def _ensure_valid_int(value: int, context: Context, bits: int):
        """
        Ensures that value is a valid int on `bits` bits.
        """
        min_val = -(1 << (bits - 1))
        max_val = (1 << (bits - 1)) - 1
        context.ensure_valid_value(
            (min_val <= value <= max_val),
            f"expected value in range [{min_val};{max_val}]",
        )



import pytest

from starknet_py.serialization.tuple_dataclass import TupleDataclass


def test_wrapped_named_tuple():
    input_dict = {
        "first": 1,
        "second": 2,
        "third": {"key": "value"},
    }
    input_tuple = tuple(input_dict.values())

    result = TupleDataclass.from_dict(input_dict)

    # __eq__ check
    assert input_tuple == result
    assert result == input_tuple
    assert result == TupleDataclass.from_dict(input_dict, name="OtherNAme")

    assert result.as_tuple() == input_tuple
    assert result.as_dict() == input_dict
    assert result._asdict() == input_dict
    assert (result[0], result[1], result[2]) == input_tuple
    assert (result.first, result.second, result.third) == input_tuple
    assert str(result) == "TupleDataclass(first=1, second=2, third={'key': 'value'})"
    assert repr(result) == "TupleDataclass(first=1, second=2, third={'key': 'value'})"

    result = TupleDataclass.from_dict(input_dict, name="CustomClass")
    assert str(result) == "CustomClass(first=1, second=2, third={'key': 'value'})"
    assert repr(result) == "CustomClass(first=1, second=2, third={'key': 'value'})"

    with pytest.raises(
        AttributeError, match="object has no attribute 'unknown_attribute'"
    ):
        result.unknown_attribute()



import pytest

from starknet_py.net.client_models import Call
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
@pytest.mark.skip  # this test is a part of migration guide
async def test_account_comparison(gateway_account, map_contract):
    address = map_contract.address
    key = 0x111
    account = gateway_account
    account_client = gateway_account

    # docs-1: start
    # Inspecting storage

    await account_client.get_storage_at(contract_address=address, key=key)

    # becomes

    await account.client.get_storage_at(contract_address=address, key=key)
    # docs-1: end

    call = Call(to_addr=0x1, selector=0x1234, calldata=[])
    resource_bounds = MAX_RESOURCE_BOUNDS

    # docs-2: start
    # Sending transactions

    tx = await account_client.sign_invoke_v3(call, resource_bounds)
    await account_client.send_transaction(tx)

    # becomes

    tx = await account.sign_invoke_v3(call, resource_bounds=resource_bounds)
    # Note that resource_bounds is now keyword-only argument
    await account.client.send_transaction(tx)
    # docs-2: end

    # docs-3: start
    # Using execute method

    await account_client.execute_v3(call, resource_bounds=resource_bounds)

    # becomes

    await account.execute_v3(call, resource_bounds=resource_bounds)
    # docs-3: end



from __future__ import annotations

import dataclasses
import json
from collections import OrderedDict, defaultdict
from typing import DefaultDict, Dict, List, Optional, Tuple, TypeVar, Union, cast

from marshmallow import EXCLUDE

from starknet_py.abi.v2 import Abi
from starknet_py.abi.v2.schemas import ContractAbiEntrySchema
from starknet_py.abi.v2.shape import (
    CONSTRUCTOR_ENTRY,
    ENUM_ENTRY,
    EVENT_ENTRY,
    FUNCTION_ENTRY,
    IMPL_ENTRY,
    INTERFACE_ENTRY,
    L1_HANDLER_ENTRY,
    STRUCT_ENTRY,
    ConstructorDict,
    EventDict,
    EventEnumVariantDict,
    EventStructMemberDict,
    FunctionDict,
    ImplDict,
    InterfaceDict,
    TypedParameterDict,
)
from starknet_py.cairo.data_types import CairoType, EnumType, EventType, StructType
from starknet_py.cairo.v2.type_parser import TypeParser


class AbiParsingError(ValueError):
    """
    Error raised when something wrong goes during abi parsing.
    """


class AbiParser:
    """
    Utility class for parsing abi into a dataclass.
    """

    # Entries from ABI grouped by entry type
    _grouped: DefaultDict[str, List[Dict]]
    # lazy init property
    _type_parser: Optional[TypeParser] = None

    def __init__(self, abi_list: List[Dict]):
        """
        Abi parser constructor. Ensures that abi satisfies the abi schema.

        :param abi_list: Contract's ABI as a list of dictionaries.
        """
        abi = [
            ContractAbiEntrySchema().load(entry, unknown=EXCLUDE) for entry in abi_list
        ]
        grouped = defaultdict(list)
        for entry in abi:
            assert isinstance(entry, dict)
            grouped[entry["type"]].append(entry)

        self._grouped = grouped

    def parse(self) -> Abi:
        """
        Parse abi provided to constructor and return it as a dataclass. Ensures that there are no cycles in the abi.

        :raises: AbiParsingError: on any parsing error.
        :return: Abi dataclass.
        """
        structures, enums = self._parse_structures_and_enums()
        events_dict = cast(
            Dict[str, EventDict],
            AbiParser._group_by_entry_name(
                self._grouped[EVENT_ENTRY], "defined events"
            ),
        )

        events: Dict[str, EventType] = {}
        for name, event in events_dict.items():
            events[name] = self._parse_event(event)
            assert self._type_parser is not None
            self._type_parser.add_defined_type(events[name])

        functions_dict = cast(
            Dict[str, FunctionDict],
            AbiParser._group_by_entry_name(
                self._grouped[FUNCTION_ENTRY], "defined functions"
            ),
        )
        interfaces_dict = cast(
            Dict[str, InterfaceDict],
            AbiParser._group_by_entry_name(
                self._grouped[INTERFACE_ENTRY], "defined interfaces"
            ),
        )
        impls_dict = cast(
            Dict[str, ImplDict],
            AbiParser._group_by_entry_name(self._grouped[IMPL_ENTRY], "defined impls"),
        )
        l1_handlers_dict = cast(
            Dict[str, FunctionDict],
            AbiParser._group_by_entry_name(
                self._grouped[L1_HANDLER_ENTRY], "defined L1 handlers"
            ),
        )
        constructors = self._grouped[CONSTRUCTOR_ENTRY]

        if len(constructors) > 1:
            raise AbiParsingError("Constructor in ABI must be defined at most once.")

        return Abi(
            defined_structures=structures,
            defined_enums=enums,
            constructor=(
                self._parse_constructor(cast(ConstructorDict, constructors[0]))
                if constructors
                else None
            ),
            l1_handler={
                name: self._parse_function(entry)
                for name, entry in l1_handlers_dict.items()
            },
            functions={
                name: self._parse_function(entry)
                for name, entry in functions_dict.items()
            },
            events=events,
            interfaces={
                name: self._parse_interface(entry)
                for name, entry in interfaces_dict.items()
            },
            implementations={
                name: self._parse_impl(entry) for name, entry in impls_dict.items()
            },
        )

    @property
    def type_parser(self) -> TypeParser:
        if self._type_parser:
            return self._type_parser

        raise RuntimeError("Tried to get type_parser before it was set.")

    def _parse_structures_and_enums(
        self,
    ) -> Tuple[Dict[str, StructType], Dict[str, EnumType]]:
        structs_dict = AbiParser._group_by_entry_name(
            self._grouped[STRUCT_ENTRY], "defined structures"
        )
        enums_dict = AbiParser._group_by_entry_name(
            self._grouped[ENUM_ENTRY], "defined enums"
        )

        # Contains sorted members of the struct
        struct_members: Dict[str, List[TypedParameterDict]] = {}
        structs: Dict[str, StructType] = {}

        # Contains sorted members of the enum
        enum_members: Dict[str, List[TypedParameterDict]] = {}
        enums: Dict[str, EnumType] = {}

        # Example problem (with a simplified json structure):
        # [{name: User, fields: {id: Uint256}}, {name: "Uint256", ...}]
        # User refers to Uint256 even though it is not known yet (will be parsed next).
        # This is why it is important to create the structure types first. This way other types can already refer to
        # them when parsing types, even thought their fields are not filled yet.
        # At the end we will mutate those structures to contain the right fields. An alternative would be to use
        # topological sorting with an additional "unresolved type", so this flow is much easier.
        for name, struct in structs_dict.items():
            structs[name] = StructType(name, OrderedDict())
            struct_members[name] = struct["members"]

        for name, enum in enums_dict.items():
            enums[name] = EnumType(name, OrderedDict())
            enum_members[name] = enum["variants"]

        # Now parse the types of members and save them.
        defined_structs_enums: Dict[str, Union[StructType, EnumType]] = dict(structs)
        defined_structs_enums.update(enums)

        self._type_parser = TypeParser(defined_structs_enums)  # pyright: ignore
        for name, struct in structs.items():
            members = self._parse_members(
                cast(List[TypedParameterDict], struct_members[name]),
                f"members of structure '{name}'",
            )
            struct.types.update(members)
        for name, enum in enums.items():
            members = self._parse_members(
                cast(List[TypedParameterDict], enum_members[name]),
                f"members of enum '{name}'",
            )
            enum.variants.update(members)

        # All types have their members assigned now

        self._check_for_cycles(defined_structs_enums)

        return structs, enums

    @staticmethod
    def _check_for_cycles(structs: Dict[str, Union[StructType, EnumType]]):
        # We want to avoid creating our own cycle checker as it would make it more complex. json module has a built-in
        # checker for cycles.
        try:
            _to_json(structs)
        except ValueError as err:
            raise AbiParsingError(err) from ValueError

    def _parse_function(self, function: FunctionDict) -> Abi.Function:
        return Abi.Function(
            name=function["name"],
            inputs=self._parse_members(function["inputs"], function["name"]),
            outputs=list(
                self.type_parser.parse_inline_type(param["type"])
                for param in function["outputs"]
            ),
        )

    def _parse_constructor(self, constructor: ConstructorDict) -> Abi.Constructor:
        return Abi.Constructor(
            name=constructor["name"],
            inputs=self._parse_members(constructor["inputs"], constructor["name"]),
        )

    def _parse_event(self, event: EventDict) -> EventType:
        members_ = event.get("members", event.get("variants"))
        assert isinstance(members_, list)
        return EventType(
            name=event["name"],
            types=self._parse_members(
                cast(List[TypedParameterDict], members_), event["name"]
            ),
            keys=self._parse_keys(cast(List[EventStructMemberDict], members_)),
        )

    TypedParam = TypeVar(
        "TypedParam", TypedParameterDict, EventStructMemberDict, EventEnumVariantDict
    )

    def _parse_members(
        self, params: List[TypedParam], entity_name: str
    ) -> OrderedDict[str, CairoType]:
        # Without cast, it complains that 'Type "TypedParameterDict" cannot be assigned to type "T@_group_by_name"'
        members = AbiParser._group_by_entry_name(cast(List[Dict], params), entity_name)
        return OrderedDict(
            (name, self.type_parser.parse_inline_type(param["type"]))
            for name, param in members.items()
        )

    def _parse_keys(self, params: List[EventStructMemberDict]) -> List[str]:
        return [param["name"] for param in params if param["kind"] == "key"]

    def _parse_interface(self, interface: InterfaceDict) -> Abi.Interface:
        return Abi.Interface(
            name=interface["name"],
            items=OrderedDict(
                (entry["name"], self._parse_function(entry))
                for entry in interface["items"]
            ),
        )

    @staticmethod
    def _parse_impl(impl: ImplDict) -> Abi.Impl:
        return Abi.Impl(
            name=impl["name"],
            interface_name=impl["interface_name"],
        )

    @staticmethod
    def _group_by_entry_name(
        dicts: List[Dict], entity_name: str
    ) -> OrderedDict[str, Dict]:
        grouped = OrderedDict()
        for entry in dicts:
            name = entry["name"]
            if name in grouped:
                raise AbiParsingError(
                    f"Name '{name}' was used more than once in {entity_name}."
                )
            grouped[name] = entry
        return grouped


def _to_json(value):
    class DataclassSupportingEncoder(json.JSONEncoder):
        def default(self, o):
            # Dataclasses are not supported by json. Additionally, dataclasses.asdict() works recursively and doesn't
            # check for cycles, so we need to flatten dataclasses (by ONE LEVEL) ourselves.
            if dataclasses.is_dataclass(o):
                return tuple(getattr(o, field.name) for field in dataclasses.fields(o))
            return super().default(o)

    return json.dumps(value, cls=DataclassSupportingEncoder)



from starknet_py.net.full_node_client import FullNodeClient


def test_synchronous_full_node_client(
    client,
    map_class_hash,  # pylint: disable=unused-argument
):
    # pylint: disable=unused-variable
    fixture_client = client

    # docs: start
    node_url = "https://your.node.url"
    client = FullNodeClient(node_url=node_url)
    # docs: end

    client = fixture_client

    # docs: start
    call_result = client.get_block_sync(block_number=1)
    # docs: end



import pytest


@pytest.mark.asyncio
async def test_create_call_from_contract(map_contract, account):
    # pylint: disable=import-outside-toplevel
    contract = map_contract

    client = account.client
    res = await map_contract.functions["put"].invoke_v3(
        key=1234, value=9999, auto_estimate=True
    )
    await res.wait_for_acceptance()

    # docs: start
    from starknet_py.net.client_models import Call

    # Prepare a call through Contract
    call = contract.functions["get"].prepare_invoke_v3(key=1234)
    assert issubclass(type(call), Call)

    # Use call directly through Client
    result = await client.call_contract(call)
    # docs: end

    assert result[0] == 9999



import pytest_asyncio

from starknet_py.devnet_utils.devnet_client import DevnetClient
from starknet_py.net.account.account import Account
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.models import StarknetChainId
from starknet_py.net.signer.key_pair import KeyPair


@pytest_asyncio.fixture(scope="package")
async def account_forked_devnet(
    devnet_client_fork_mode: DevnetClient,
) -> BaseAccount:
    predeployed_account_info = (
        await devnet_client_fork_mode.get_predeployed_accounts()
    )[0]
    return Account(
        address=predeployed_account_info.address,
        client=devnet_client_fork_mode,
        key_pair=KeyPair.from_private_key(predeployed_account_info.private_key),
        chain=StarknetChainId.SEPOLIA,
    )


@pytest_asyncio.fixture(scope="package")
async def account_to_impersonate(devnet_client_fork_mode: DevnetClient) -> BaseAccount:
    """
    Creates an account instance for impersonation.

    :param address: address from Sepolia testnet that is not in the local state,
                    so it can be impersonated.
    """
    account = Account(
        address="0x043abaa073c768ebf039c0c4f46db9acc39e9ec165690418060a652aab39e7d8",
        client=devnet_client_fork_mode,
        key_pair=KeyPair(private_key="0x1", public_key="0x1"),
        chain=StarknetChainId.SEPOLIA,
    )
    await devnet_client_fork_mode.mint(account.address, int(1e40), "FRI")

    return account



import pytest

from starknet_py.net.models.address import parse_address


@pytest.mark.parametrize(
    "input_addr, output",
    [(123, 123), ("859", 2137), ("0x859", 2137)],
)
def test_parse_address(input_addr, output):
    assert parse_address(input_addr) == output


def test_parse_invalid_address():
    with pytest.raises(TypeError, match="address format"):
        # Ignore typing, because it is an error check (float can't be passed here)
        # noinspection PyTypeChecker
        parse_address(0.22)  # pyright: ignore



from typing import Union

AddressRepresentation = Union[int, str]
Address = int


def parse_address(value: AddressRepresentation) -> Address:
    if isinstance(value, int):
        return value

    try:
        return int(value, 16)
    except TypeError as t_err:
        raise TypeError("Invalid address format.") from t_err



import json

import pytest

from starknet_py.net.client_models import SierraContractClass
from starknet_py.net.udc_deployer.deployer import _get_random_salt
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import load_contract


@pytest.mark.asyncio
async def test_cairo1_contract(
    account, sierra_minimal_compiled_contract_and_class_hash
):
    # pylint: disable=too-many-locals
    # pylint: disable=import-outside-toplevel
    (
        compiled_contract,
        compiled_class_hash,
    ) = sierra_minimal_compiled_contract_and_class_hash

    contract_compiled_casm = load_contract(
        "MinimalContract",
    )["casm"]

    # docs: start
    from starknet_py.common import create_casm_class
    from starknet_py.hash.casm_class_hash import compute_casm_class_hash

    # contract_compiled_casm is a string containing the content of the starknet-sierra-compile (.casm file)
    casm_class = create_casm_class(contract_compiled_casm)

    # Compute Casm class hash
    casm_class_hash = compute_casm_class_hash(casm_class)
    # docs: end

    assert casm_class_hash == compiled_class_hash

    # docs: start

    # Create Declare v3 transaction
    declare_v3_transaction = await account.sign_declare_v3(
        # compiled_contract is a string containing the content of the starknet-compile (.json file)
        compiled_contract=compiled_contract,
        compiled_class_hash=casm_class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    # Send transaction
    resp = await account.client.declare(transaction=declare_v3_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    sierra_class_hash = resp.class_hash
    # docs: end

    assert sierra_class_hash != 0

    # START OF DEPLOY SECTION
    calldata = []
    salt = _get_random_salt()
    abi = json.loads(compiled_contract)["abi"]
    # docs-deploy: start
    from starknet_py.net.udc_deployer.deployer import Deployer

    # Use Universal Deployer Contract (UDC) to deploy the Cairo1 contract
    deployer = Deployer()

    # Create a ContractDeployment, optionally passing salt and calldata
    contract_deployment = deployer.create_contract_deployment(
        class_hash=sierra_class_hash,
        abi=abi,
        cairo_version=1,
        calldata=calldata,
        salt=salt,
    )

    res = await account.execute_v3(
        calls=contract_deployment.call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await account.client.wait_for_tx(res.transaction_hash)

    # The contract has been deployed and can be found at contract_deployment.address
    # docs-deploy: end

    assert isinstance(contract_deployment.address, int)
    assert contract_deployment.address != 0

    compiled_class = await account.client.get_class_by_hash(
        class_hash=sierra_class_hash
    )
    assert isinstance(compiled_class, SierraContractClass)



import pytest

from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping


@pytest.mark.asyncio
async def test_simple_deploy(account, hello_starknet_class_hash, hello_starknet_abi):
    # pylint: disable=import-outside-toplevel
    # docs: start
    from starknet_py.contract import Contract

    # docs: end

    class_hash = hello_starknet_class_hash
    abi = hello_starknet_abi

    # docs: start
    # To deploy contract just use `Contract.deploy_contract_v3` method
    # Note that class_hash and abi of the contract must be known

    # If constructor of the contract requires arguments, pass constructor_args parameter
    constructor_args = {"value": 10}

    # docs: end
    constructor_args = None

    # docs: start
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    deploy_result = await Contract.deploy_contract_v3(
        account=account,
        class_hash=class_hash,
        abi=abi,  # abi is optional
        constructor_args=constructor_args,
        resource_bounds=resource_bounds,
    )

    # `Contract.deploy_contract_v3` method has an optional parameter
    # `deployer_address` that needs to be specified when using other network than mainnet or sepolia
    # Read more about it in the API section

    # Wait for the transaction
    await deploy_result.wait_for_acceptance()

    # To interact with just deployed contract get its instance from the deploy_result
    contract = deploy_result.deployed_contract

    # Now, any of the contract functions can be called
    # docs: end

    assert contract.address != 0



from typing import Any, Optional, cast

from marshmallow import ValidationError, fields, post_load

from starknet_py.net.client_models import Transaction, TransactionStatusWithoutL1
from starknet_py.net.client_utils import _to_rpc_felt
from starknet_py.net.schemas.common import Felt, TransactionStatusWithoutL1Field
from starknet_py.net.schemas.rpc.block import BlockHeaderSchema
from starknet_py.net.schemas.rpc.event import EmittedEventWithFinalitySchema
from starknet_py.net.schemas.rpc.transactions import (
    TransactionReceiptWithBlockInfoSchema,
    TransactionStatusResponseSchema,
    TypesOfTransactionsSchema,
)
from starknet_py.net.websockets.models import (
    NewEventsNotification,
    NewHeadsNotification,
    NewTransactionNotification,
    NewTransactionNotificationResult,
    NewTransactionReceiptsNotification,
    NewTransactionStatus,
    ReorgData,
    ReorgNotification,
    TransactionStatusNotification,
)
from starknet_py.utils.schema import Schema


class NewHeadsNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(BlockHeaderSchema(), data_key="result", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> NewHeadsNotification:
        return NewHeadsNotification(**data)


class NewEventsNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(
        EmittedEventWithFinalitySchema(), data_key="result", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> NewEventsNotification:
        return NewEventsNotification(**data)


class NewTransactionStatusSchema(Schema):
    transaction_hash = Felt(data_key="transaction_hash", required=True)
    status = fields.Nested(
        TransactionStatusResponseSchema(), data_key="status", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> NewTransactionStatus:
        return NewTransactionStatus(**data)


class TransactionStatusNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(
        NewTransactionStatusSchema(), data_key="result", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> TransactionStatusNotification:
        return TransactionStatusNotification(**data)


class PendingTransactionsNotificationResultField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if isinstance(value, int):
            return _to_rpc_felt(value)
        elif isinstance(value, Transaction):
            return TypesOfTransactionsSchema().dump(value)
        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}"
        )

    def _deserialize(
        self,
        value,
        attr,
        data,
        **kwargs,
    ):
        if isinstance(value, str):
            return Felt().deserialize(value, attr, data, **kwargs)
        elif isinstance(value, dict):
            return TypesOfTransactionsSchema().load(value)
        raise ValidationError(
            f"Invalid value provided for {self.__class__.__name__}: {value}"
        )


class ReorgDataSchema(Schema):
    starting_block_hash = Felt(data_key="starting_block_hash", required=True)
    starting_block_number = fields.Integer(
        data_key="starting_block_number", required=True, validate=lambda x: x >= 0
    )
    ending_block_hash = Felt(data_key="ending_block_hash", required=True)
    ending_block_number = fields.Integer(
        data_key="ending_block_number", required=True, validate=lambda x: x >= 0
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ReorgData:
        return ReorgData(**data)


class ReorgNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(ReorgDataSchema(), data_key="result", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> ReorgNotification:
        return ReorgNotification(**data)


class NewTransactionReceiptsNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(
        TransactionReceiptWithBlockInfoSchema(), data_key="result", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> NewTransactionReceiptsNotification:
        return NewTransactionReceiptsNotification(**data)


class TypesOfTransactionWithFinalitySchema(TypesOfTransactionsSchema):
    def load(self, data, *, many=None, partial=None, unknown=None, **kwargs):
        if not isinstance(data, dict):
            raise ValidationError(
                "Invalid payload for transaction with finality_status."
            )

        finality_status_field = TransactionStatusWithoutL1Field(
            data_key="finality_status", required=True
        )
        finality_status = finality_status_field.deserialize(
            data.get("finality_status"),
            attr="finality_status",
            data=data,
        )
        finality_status = cast(TransactionStatusWithoutL1, finality_status)

        tx_payload = dict(data)
        tx_payload.pop("finality_status", None)
        tx_obj = super().load(tx_payload, many=many, partial=partial, unknown=unknown)
        tx_obj = cast(Transaction, tx_obj)

        return NewTransactionNotificationResult(
            transaction=tx_obj, finality_status=finality_status
        )


class NewTransactionNotificationSchema(Schema):
    subscription_id = fields.Str(data_key="subscription_id", required=True)
    result = fields.Nested(
        TypesOfTransactionWithFinalitySchema(), data_key="result", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> NewTransactionNotification:
        return NewTransactionNotification(**data)



from collections import OrderedDict

import pytest

from starknet_py.cairo.data_types import (
    ArrayType,
    FeltType,
    NamedTupleType,
    StructType,
    TupleType,
)
from starknet_py.cairo.type_parser import TypeParser, UnknownCairoTypeError


@pytest.mark.parametrize(
    "type_string, expected",
    [
        ("felt", FeltType()),
        ("felt*", ArrayType(FeltType())),
        ("(felt, felt, felt, felt)", TupleType([FeltType()] * 4)),
        (
            "(low: felt, high: felt)",
            NamedTupleType(OrderedDict(low=FeltType(), high=FeltType())),
        ),
        ("()", TupleType([])),
        (
            "(a: felt, b: (felt, (felt*, felt)))",
            NamedTupleType(
                OrderedDict(
                    a=FeltType(),
                    b=TupleType(
                        [
                            FeltType(),
                            TupleType(
                                [
                                    ArrayType(FeltType()),
                                    FeltType(),
                                ]
                            ),
                        ]
                    ),
                )
            ),
        ),
    ],
)
def test_parse_without_defined_types(type_string, expected):
    parsed = TypeParser({}).parse_inline_type(type_string)
    assert parsed == expected


uint256_type = StructType("Uint256", OrderedDict(low=FeltType(), high=FeltType()))
wrapped_felt_type = StructType("WrappedFelt", OrderedDict(value=FeltType()))


@pytest.mark.parametrize(
    "type_string, expected",
    [
        ("Uint256", uint256_type),
        ("Uint256*", ArrayType(uint256_type)),
        ("(Uint256, WrappedFelt)", TupleType([uint256_type, wrapped_felt_type])),
        (
            "(a: Uint256, b: (WrappedFelt, (felt*, WrappedFelt*)))",
            NamedTupleType(
                OrderedDict(
                    a=uint256_type,
                    b=TupleType(
                        [
                            wrapped_felt_type,
                            TupleType(
                                [
                                    ArrayType(FeltType()),
                                    ArrayType(wrapped_felt_type),
                                ]
                            ),
                        ]
                    ),
                )
            ),
        ),
    ],
)
def test_parse_with_defined_types(type_string, expected):
    types = {
        "Uint256": uint256_type,
        "WrappedFelt": wrapped_felt_type,
    }
    parsed = TypeParser(types).parse_inline_type(type_string)
    assert parsed == expected


def test_code_offset():
    # cairo-lang parser treats codeoffset specially, but we just want to treat it as a type defined by the user.

    # codeoffset is not defined
    with pytest.raises(UnknownCairoTypeError) as err_info:
        TypeParser({}).parse_inline_type("codeoffset")

    assert err_info.value.type_name == "codeoffset"

    # codeoffset is defined
    struct = StructType("codeoffset", OrderedDict(value=FeltType()))
    parsed = TypeParser({"codeoffset": struct}).parse_inline_type("codeoffset")
    assert parsed == struct


def test_missing_type():
    with pytest.raises(
        UnknownCairoTypeError, match="Type 'Uint256' is not defined"
    ) as err_info:
        TypeParser({}).parse_inline_type("Uint256")

    assert err_info.value.type_name == "Uint256"


def test_names_not_matching():
    with pytest.raises(
        ValueError, match="Keys must match name of type, 'OtherName' != 'Uint256'."
    ):
        TypeParser({"OtherName": uint256_type})



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, OrderedDict, Union

from starknet_py.cairo.data_types import CairoType, EnumType, EventType, StructType


@dataclass
class Abi:
    """
    Dataclass representing class abi. Contains parsed functions, enums, events and structures.
    """

    # pylint: disable=too-many-instance-attributes

    @dataclass
    class Function:
        """
        Dataclass representing function's abi.
        """

        name: str
        inputs: OrderedDict[str, CairoType]
        outputs: List[CairoType]

    @dataclass
    class Constructor:
        """
        Dataclass representing constructor's abi.
        """

        name: str
        inputs: OrderedDict[str, CairoType]

    @dataclass
    class EventStruct:
        """
        Dataclass representing struct event's abi.
        """

        name: str
        members: OrderedDict[str, CairoType]

    @dataclass
    class EventEnum:
        """
        Dataclass representing enum event's abi.
        """

        name: str
        variants: OrderedDict[str, CairoType]

    Event = Union[EventStruct, EventEnum]

    @dataclass
    class Interface:
        """
        Dataclass representing an interface.
        """

        name: str
        items: OrderedDict[
            str, Abi.Function
        ]  # Only functions can be defined in the interface

    @dataclass
    class Impl:
        """
        Dataclass representing an impl.
        """

        name: str
        interface_name: str

    defined_structures: Dict[
        str, StructType
    ]  #: Abi of structures defined by the class.
    defined_enums: Dict[str, EnumType]  #: Abi of enums defined by the class.
    functions: Dict[str, Function]  #: Functions defined by the class.
    events: Dict[str, EventType]  #: Events defined by the class
    constructor: Optional[
        Constructor
    ]  #: Contract's constructor. It is None if class doesn't define one.
    l1_handler: Optional[
        Dict[str, Function]
    ]  #: Handlers of L1 messages. It is None if class doesn't define one.
    interfaces: Dict[str, Interface]
    implementations: Dict[str, Impl]



import pytest

from starknet_py.constants import FIELD_PRIME
from starknet_py.serialization import FeltSerializer
from starknet_py.serialization.data_serializers.non_zero_serializer import (
    NonZeroSerializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (NonZeroSerializer(UintSerializer(128)), 123, [123]),
        (NonZeroSerializer(UintSerializer(256)), 1, [1, 0]),
        (NonZeroSerializer(FeltSerializer()), 10, [10]),
        (NonZeroSerializer(FeltSerializer()), FIELD_PRIME - 1, [FIELD_PRIME - 1]),
    ],
)
def test_valid_values(serializer, value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (NonZeroSerializer(UintSerializer(128)), 0, [0]),
        (NonZeroSerializer(UintSerializer(256)), 0, [0, 0]),
        (NonZeroSerializer(FeltSerializer()), 0, [0]),
    ],
)
def test_invalid_values(serializer, value, serialized_value):
    with pytest.raises(ValueError, match="expected value to be non-zero"):
        serializer.deserialize(serialized_value)
        serializer.serialize(value)



import pytest

from starknet_py.contract import Contract, DeclareResult, DeployResult
from starknet_py.net.account.base_account import BaseAccount


@pytest.mark.parametrize("param", ["_account", "class_hash", "compiled_contract"])
def test_declare_result_post_init(param, account):
    kwargs = {
        "_account": account,
        "class_hash": 0,
        "compiled_contract": "",
    }
    del kwargs[param]

    with pytest.raises(ValueError, match=f"Argument {param} can't be None."):
        _ = DeclareResult(hash=0, _client=account.client, **kwargs)


def test_deploy_result_post_init(client):
    with pytest.raises(ValueError, match="Argument deployed_contract can't be None."):
        _ = DeployResult(
            hash=0,
            _client=client,
        )


def test_contract_raises_on_incorrect_provider_type():
    with pytest.raises(ValueError, match="Argument provider is not of accepted type."):
        Contract(address=0x1, abi=[], provider=1)  # pyright: ignore


def test_contract_create_with_base_account(account):
    contract = Contract(address=0x1, abi=[], provider=account)
    assert isinstance(contract.account, BaseAccount)
    assert contract.account == account
    assert contract.client == account.client


def test_contract_create_with_client(client):
    contract = Contract(address=0x1, abi=[], provider=client)
    assert contract.account is None
    assert contract.client == client


def test_throws_on_wrong_abi(account):
    with pytest.raises(
        ValueError, match="Make sure valid ABI is used to create a Contract instance"
    ):
        Contract(
            address=0x1,
            abi=[
                {
                    "type": "function",
                    "name": "empty",
                    "inputs": "",  # inputs should be a list
                }
            ],
            provider=account,
            cairo_version=1,
        )



from collections import OrderedDict

import pytest

from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.struct_serializer import (
    StructSerializer,
)

felt_array_serializer = ArraySerializer(FeltSerializer())


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (
            StructSerializer(OrderedDict(x=FeltSerializer(), y=FeltSerializer())),
            # Reversed order, serializer should read them properly
            {"y": 2, "x": 1},
            [1, 2],
        ),
        (
            StructSerializer(
                OrderedDict(
                    inner=StructSerializer(
                        OrderedDict(x=FeltSerializer(), y=felt_array_serializer)
                    )
                )
            ),
            {"inner": {"x": 22, "y": [38]}},
            [22, 1, 38],
        ),
    ],
)
def test_valid_values(serializer, value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value



from dataclasses import dataclass

from starknet_py.contract import Contract
from starknet_py.net.account.account import Account
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.utils import (
    AccountToBeDeployedDetails,
    get_deploy_account_transaction,
)


@dataclass
class PreparedNetworkData:
    # pylint: disable=too-many-instance-attributes
    class_hash: int
    contract_address: int
    contract_address_2: int
    invoke_transaction_hash: int
    block_with_invoke_number: int
    declare_transaction_hash: int
    block_with_declare_number: int
    block_with_declare_hash: int
    deploy_account_transaction_hash: int
    block_with_deploy_account_number: int
    block_with_deploy_account_hash: int


async def prepare_net_for_tests(
    account: Account,
    deploy_account_details: AccountToBeDeployedDetails,
    transaction_hash: int,
    contract: Contract,
    contract_2: Contract,
    declare_class_hash: int,
) -> PreparedNetworkData:
    # pylint: disable=too-many-locals

    declare_receipt = await account.client.get_transaction_receipt(transaction_hash)
    block_with_declare_number = declare_receipt.block_number
    block_with_declare_hash = declare_receipt.block_hash

    invoke_res = await contract.functions["increase_balance"].invoke_v3(
        amount=1777, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await invoke_res.wait_for_acceptance()

    invoke_res_2 = await contract_2.functions["increase_balance"].invoke_v3(
        amount=1777, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await invoke_res_2.wait_for_acceptance()

    block_with_invoke_number = (
        await account.client.get_transaction_receipt(invoke_res.hash)
    ).block_number
    block_with_invoke_number_2 = (
        await account.client.get_transaction_receipt(invoke_res_2.hash)
    ).block_number

    address, key_pair, salt, class_hash = deploy_account_details
    deploy_account_tx = await get_deploy_account_transaction(
        address=address,
        key_pair=key_pair,
        salt=salt,
        class_hash=class_hash,
        client=account.client,
    )
    deploy_account_result = await account.client.deploy_account(deploy_account_tx)
    await account.client.wait_for_tx(deploy_account_result.transaction_hash)

    declare_account_receipt = await account.client.get_transaction_receipt(
        deploy_account_result.transaction_hash
    )
    block_with_deploy_account_number = declare_account_receipt.block_number
    block_with_deploy_account_hash = declare_account_receipt.block_hash

    assert block_with_invoke_number is not None
    assert block_with_invoke_number_2 is not None
    assert block_with_declare_number is not None
    assert block_with_declare_hash is not None
    assert block_with_deploy_account_number is not None
    assert block_with_deploy_account_hash is not None

    return PreparedNetworkData(
        class_hash=declare_class_hash,
        contract_address=contract.address,
        contract_address_2=contract_2.address,
        invoke_transaction_hash=invoke_res.hash,
        block_with_invoke_number=block_with_invoke_number,
        declare_transaction_hash=transaction_hash,
        block_with_declare_number=block_with_declare_number,
        block_with_declare_hash=block_with_declare_hash,
        deploy_account_transaction_hash=deploy_account_result.transaction_hash,
        block_with_deploy_account_hash=block_with_deploy_account_hash,
        block_with_deploy_account_number=block_with_deploy_account_number,
    )



from starknet_py.constants import FIELD_PRIME
from starknet_py.net.signer.key_pair import KeyPair


def test_generate_key_pair():
    key_pair = KeyPair.generate()
    hex_private_key = hex(key_pair.private_key)[2:]
    padded_hex_private_key = hex_private_key.zfill(64)
    assert len(padded_hex_private_key) == 64
    assert key_pair.private_key < FIELD_PRIME
    assert key_pair.public_key < FIELD_PRIME



import pytest

from starknet_py.serialization._context import SerializationContext
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException


@pytest.mark.parametrize(
    "initial_exception, wrapped_class",
    [
        (ValueError("Test"), InvalidValueException),
        (TypeError("Test"), InvalidTypeException),
    ],
)
def test_if_exceptions_are_wrapped(initial_exception, wrapped_class):
    with pytest.raises(wrapped_class, match="Error: Test"):
        with SerializationContext.create():
            raise initial_exception



from typing import Literal, Union

from starknet_py.constants import ETH_FEE_CONTRACT_ADDRESS

MAINNET = "mainnet"
SEPOLIA = "sepolia"
SEPOLIA_INTEGRATION = "sepolia_integration"

PredefinedNetwork = Literal["mainnet", "sepolia", "sepolia_integration"]

Network = Union[PredefinedNetwork, str]


def default_token_address_for_network(net: Network) -> str:
    if net not in [MAINNET, SEPOLIA, SEPOLIA_INTEGRATION]:
        raise ValueError(
            "Argument token_address must be specified when using a custom net address"
        )

    return ETH_FEE_CONTRACT_ADDRESS



from typing import AsyncGenerator

import pytest
import pytest_asyncio

from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.websockets.websocket_client import WebsocketClient


@pytest.fixture(name="client", scope="package")
def create_full_node_client(devnet: str) -> FullNodeClient:
    return FullNodeClient(node_url=devnet + "/rpc")


@pytest.fixture(name="client_fork_mode", scope="package")
def create_full_node_client_fork_mode(devnet_forking_mode: str) -> FullNodeClient:
    return FullNodeClient(node_url=devnet_forking_mode + "/rpc")


@pytest_asyncio.fixture(scope="package")
async def websocket_client(devnet_ws: str) -> AsyncGenerator[WebsocketClient, None]:
    """
    Connects `WebsocketClient` client to devnet, returns its instance and disconnects after the tests.
    """
    client = WebsocketClient(devnet_ws)

    await client.connect()
    yield client
    await client.disconnect()



import lark

from starknet_py.cairo.deprecated_parse.cairo_types import CairoType
from starknet_py.cairo.deprecated_parse.parser_transformer import ParserTransformer

CAIRO_EBNF = """
    %import common.WS_INLINE
    %ignore WS_INLINE

    IDENTIFIER: /[a-zA-Z_][a-zA-Z_0-9]*/
    _DBL_STAR: "**"
    COMMA: ","

    ?type: non_identifier_type
        | identifier             -> type_struct

    comma_separated{item}: item? (COMMA item)* COMMA?

    named_type: identifier (":" type)? | non_identifier_type
    non_identifier_type: "felt"                                         -> type_felt
                    | "codeoffset"                                   -> type_codeoffset
                    | type "*"                                       -> type_pointer
                    | type _DBL_STAR                                 -> type_pointer2
                    | "(" comma_separated{named_type} ")" -> type_tuple

    identifier: IDENTIFIER ("." IDENTIFIER)*
"""


def parse(code: str) -> CairoType:
    """
    Parses the given string and returns a CairoType.
    """

    grammar = CAIRO_EBNF

    grammar_parser = lark.Lark(
        grammar=grammar,
        start=["type"],
        parser="lalr",
    )

    parsed = grammar_parser.parse(code)
    transformed = ParserTransformer().transform(parsed)

    return transformed



from enum import Enum
from typing import List

from poseidon_py.poseidon_hash import poseidon_hash, poseidon_hash_many

from starknet_py.hash.blake2s import blake2s_hash_many
from starknet_py.hash.utils import compute_hash_on_elements, pedersen_hash


class HashMethod(Enum):
    """
    Enum representing hash method.
    """

    PEDERSEN = "pedersen"
    POSEIDON = "poseidon"
    BLAKE2S = "blake2s"

    def hash(self, left: int, right: int):
        if self == HashMethod.PEDERSEN:
            return pedersen_hash(left, right)
        if self == HashMethod.POSEIDON:
            return poseidon_hash(left, right)
        if self == HashMethod.BLAKE2S:
            return blake2s_hash_many([left, right])
        raise ValueError(f"Unsupported hash method: {self}.")

    def hash_many(self, values: List[int]):
        if self == HashMethod.PEDERSEN:
            return compute_hash_on_elements(values)
        if self == HashMethod.POSEIDON:
            return poseidon_hash_many(values)
        if self == HashMethod.BLAKE2S:
            return blake2s_hash_many(values)
        raise ValueError(f"Unsupported hash method: {self}.")
